import reprlib
from copy import copy
from dataclasses import dataclass
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generator,
    Iterable,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

import numpy as np
import polars as pl
from sklearn.preprocessing import KBinsDiscretizer
from tqdm import tqdm

from eir.data_load.data_source_modules.deeplake_ops import (
    is_deeplake_dataset,
    is_deeplake_sample_missing,
    load_deeplake_dataset,
)
from eir.data_load.label_setup import (
    Labels,
    TabularFileInfo,
    al_label_transformers,
    gather_ids_from_data_source,
    gather_ids_from_tabular_file,
    get_file_path_iterator,
    set_up_train_and_valid_tabular_data,
)
from eir.experiment_io.label_transformer_io import save_transformer_set
from eir.setup import schemas
from eir.setup.schema_modules.output_schemas_survival import SurvivalOutputTypeConfig
from eir.setup.schema_modules.output_schemas_tabular import TabularOutputTypeConfig
from eir.target_setup.target_setup_utils import IdentityTransformer
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.train import Hooks


logger = get_logger(name=__name__)


def set_up_all_targets_wrapper(
    train_ids: Sequence[str],
    valid_ids: Sequence[str],
    run_folder: Path,
    output_configs: Sequence[schemas.OutputConfig],
    hooks: Optional["Hooks"],
) -> "MergedTargetLabels":
    logger.info("Setting up target labels.")

    target_labels = set_up_all_target_labels_wrapper(
        output_configs=output_configs,
        train_ids=train_ids,
        valid_ids=valid_ids,
    )
    save_transformer_set(
        transformers_per_source=target_labels.label_transformers,
        run_folder=run_folder,
    )

    return target_labels


@dataclass
class LinkedTargets:
    """
    This is specifically to track which targets should be treated as linked w.r.t.
    to missingness, e.g. in survival analysis, the event column is linked to the time
    column. While one could implicitly set this up my forcing them to have the
    same NaN values, this allows us to track this explicitly, for example when
    we filter missing outputs and target labels, we can check for this and properly
    filter the time column even though it's not an output from the model.
    """

    target_name: str
    linked_target_name: str


@dataclass
class MissingTargetsInfo:
    """
    In the case for example when we only have 1 tabular output with multiple
    targets coming from the same .csv and we have an input where we have different
    set of overall IDs (i.e., rows in the .csv) compared to the targets, we
    do not need the specific filtering of IDs later per step as all samples
    w/o inputs or targets have been filtered out in the dataset creation. We
    can track this using the all_have_same_set, which allows us to skip
    checking for the missing IDs completely before loss/metric calculation.
    """

    missing_ids_per_modality: dict[str, set[str]]
    all_have_same_set: bool


def get_missing_targets_info(
    missing_ids_per_modality: dict[str, set[str]],
) -> MissingTargetsInfo:
    if not missing_ids_per_modality:
        return MissingTargetsInfo(
            missing_ids_per_modality=missing_ids_per_modality,
            all_have_same_set=True,
        )

    sets = list(missing_ids_per_modality.values())
    all_same = len(sets) > 1 and all(s == sets[0] for s in sets[1:])

    return MissingTargetsInfo(
        missing_ids_per_modality=missing_ids_per_modality,
        all_have_same_set=all_same,
    )


def log_missing_targets_info(
    missing_targets_info: MissingTargetsInfo, all_ids: set[str]
) -> None:
    repr_formatter = reprlib.Repr()
    repr_formatter.maxset = 10

    logger.info(
        "Checking for missing target information. "
        "These will be ignored during loss and metric computation."
    )

    total_ids_count = len(all_ids)

    for modality, missing_ids in missing_targets_info.missing_ids_per_modality.items():
        missing_count = len(missing_ids)

        if missing_count == 0:
            logger.info(f"Output modality '{modality}' has no missing target IDs.")
            continue

        formatted_missing_ids = repr_formatter.repr(missing_ids)
        complete_count = total_ids_count - missing_count
        fraction_complete = (complete_count / total_ids_count) * 100

        logger.info(
            f"Missing target IDs for modality: '{modality}'\n"
            f"  - Missing IDs: {formatted_missing_ids}\n"
            f"  - Stats: Missing: {missing_count}, "
            f"Complete: {complete_count}/{total_ids_count} "
            f"({fraction_complete:.2f}% complete)\n"
        )


@dataclass
class MergedTargetLabels:
    train_labels: pl.DataFrame
    valid_labels: pl.DataFrame
    label_transformers: dict[str, al_label_transformers]
    missing_ids_per_output: MissingTargetsInfo

    @property
    def all_labels(self) -> pl.DataFrame:
        return pl.concat([self.train_labels, self.valid_labels])


def update_labels_df(
    master_df: pl.DataFrame,
    new_labels: pl.DataFrame,
    output_name: str,
) -> pl.DataFrame:
    if master_df.height == 0:
        master_df = pl.DataFrame(schema={"ID": pl.Utf8})

    rename_exprs = []
    for col in new_labels.columns:
        if col == "ID":
            rename_exprs.append(pl.col("ID"))
        else:
            new_name = f"{output_name}__{col}"
            rename_exprs.append(pl.col(col).alias(new_name))

    new_labels_renamed = new_labels.select(rename_exprs)

    if master_df.columns == ["ID"]:
        return new_labels_renamed
    else:
        return master_df.join(new_labels_renamed, on="ID", how="full", coalesce=True)


def set_up_all_target_labels_wrapper(
    output_configs: Sequence[schemas.OutputConfig],
    train_ids: Sequence[str],
    valid_ids: Sequence[str],
) -> MergedTargetLabels:
    all_ids: set[str] = set(train_ids).union(set(valid_ids))
    per_modality_missing_ids: dict[str, set[str]] = {}
    label_transformers: dict[str, Any] = {}

    train_labels_df = pl.DataFrame(schema={"ID": pl.Utf8})
    valid_labels_df = pl.DataFrame(schema={"ID": pl.Utf8})

    tabular_target_labels_info = get_tabular_target_file_infos(
        output_configs=output_configs
    )

    for output_config in output_configs:
        output_source = output_config.output_info.output_source
        output_name = output_config.output_info.output_name
        output_type = output_config.output_info.output_type
        logger.info(f"Setting up target labels for {output_name}.")

        match output_type:
            case "tabular":
                train_labels_df, valid_labels_df = process_tabular_output(
                    output_name=output_name,
                    tabular_target_labels_info=tabular_target_labels_info,
                    train_ids=train_ids,
                    valid_ids=valid_ids,
                    all_ids=all_ids,
                    label_transformers=label_transformers,
                    per_modality_missing_ids=per_modality_missing_ids,
                    train_labels_df=train_labels_df,
                    valid_labels_df=valid_labels_df,
                )
            case "sequence":
                train_labels_df, valid_labels_df = process_sequence_output(
                    output_name=output_name,
                    train_ids=train_ids,
                    valid_ids=valid_ids,
                    output_source=output_source,
                    per_modality_missing_ids=per_modality_missing_ids,
                    train_labels_df=train_labels_df,
                    valid_labels_df=valid_labels_df,
                )
            case "array" | "image":
                train_labels_df, valid_labels_df = process_array_or_image_output(
                    output_name=output_name,
                    train_ids=train_ids,
                    valid_ids=valid_ids,
                    output_config=output_config,
                    output_source=output_source,
                    per_modality_missing_ids=per_modality_missing_ids,
                    train_labels_df=train_labels_df,
                    valid_labels_df=valid_labels_df,
                )
            case "survival":
                output_type_info = output_config.output_type_info
                assert isinstance(output_type_info, SurvivalOutputTypeConfig)
                n_bins = output_type_info.num_durations
                train_labels_df, valid_labels_df = process_survival_output(
                    n_bins=n_bins,
                    output_name=output_name,
                    tabular_target_labels_info=tabular_target_labels_info,
                    train_ids=train_ids,
                    valid_ids=valid_ids,
                    all_ids=all_ids,
                    label_transformers=label_transformers,
                    per_modality_missing_ids=per_modality_missing_ids,
                    train_labels_df=train_labels_df,
                    valid_labels_df=valid_labels_df,
                )
            case _:
                raise ValueError(f"Unknown output type: '{output_type}'.")

    missing_target_info = get_missing_targets_info(
        missing_ids_per_modality=per_modality_missing_ids,
    )
    log_missing_targets_info(missing_targets_info=missing_target_info, all_ids=all_ids)

    return MergedTargetLabels(
        train_labels=train_labels_df,
        valid_labels=valid_labels_df,
        label_transformers=label_transformers,
        missing_ids_per_output=missing_target_info,
    )


def process_tabular_output(
    output_name: str,
    tabular_target_labels_info: Dict[str, Any],
    train_ids: Sequence[str],
    valid_ids: Sequence[str],
    all_ids: set[str],
    label_transformers: Dict[str, Any],
    per_modality_missing_ids: Dict[str, set[str]],
    train_labels_df: pl.DataFrame,
    valid_labels_df: pl.DataFrame,
) -> tuple[pl.DataFrame, pl.DataFrame]:
    tabular_info = tabular_target_labels_info[output_name]
    cur_labels = set_up_train_and_valid_tabular_data(
        tabular_file_info=tabular_info,
        train_ids=train_ids,
        valid_ids=valid_ids,
    )
    label_transformers[output_name] = cur_labels.label_transformers

    all_labels = cur_labels.all_labels
    cur_ids = set(all_labels.get_column("ID").to_list())
    missing_ids = all_ids.difference(cur_ids)
    per_modality_missing_ids[output_name] = missing_ids

    train_labels_df = update_labels_df(
        master_df=train_labels_df,
        new_labels=cur_labels.train_labels,
        output_name=output_name,
    )
    valid_labels_df = update_labels_df(
        master_df=valid_labels_df,
        new_labels=cur_labels.valid_labels,
        output_name=output_name,
    )

    return train_labels_df, valid_labels_df


def process_sequence_output(
    output_name: str,
    train_ids: Sequence[str],
    valid_ids: Sequence[str],
    output_source: str,
    per_modality_missing_ids: Dict[str, set[str]],
    train_labels_df: pl.DataFrame,
    valid_labels_df: pl.DataFrame,
) -> tuple[pl.DataFrame, pl.DataFrame]:
    cur_labels = set_up_delayed_target_labels(
        train_ids=train_ids,
        valid_ids=valid_ids,
        output_name=output_name,
    )

    logger.debug("Estimating missing IDs for sequence output %s.", output_name)
    missing_sequence_ids = find_sequence_output_missing_ids(
        train_ids=train_ids,
        valid_ids=valid_ids,
        output_source=output_source,
    )

    per_modality_missing_ids[output_name] = missing_sequence_ids

    train_labels_df = update_labels_df(
        master_df=train_labels_df,
        new_labels=cur_labels.train_labels,
        output_name=output_name,
    )
    valid_labels_df = update_labels_df(
        master_df=valid_labels_df,
        new_labels=cur_labels.valid_labels,
        output_name=output_name,
    )

    return train_labels_df, valid_labels_df


def process_array_or_image_output(
    output_name: str,
    train_ids: Sequence[str],
    valid_ids: Sequence[str],
    output_config: schemas.OutputConfig,
    output_source: str,
    per_modality_missing_ids: Dict[str, set[str]],
    train_labels_df: pl.DataFrame,
    valid_labels_df: pl.DataFrame,
) -> tuple[pl.DataFrame, pl.DataFrame]:
    cur_labels = set_up_file_target_labels(
        train_ids=train_ids,
        valid_ids=valid_ids,
        output_config=output_config,
    )

    logger.debug("Estimating missing IDs for array output %s.", output_name)
    cur_missing_ids = gather_torch_null_missing_ids(
        labels=cur_labels.all_labels,
        output_name=output_name,
    )
    per_modality_missing_ids[output_name] = cur_missing_ids

    is_deeplake = is_deeplake_dataset(data_source=output_source)
    col_name = f"{output_name}__{output_name}"

    polars_dtype: Type[pl.Int64] | Type[pl.Utf8]
    if is_deeplake:
        polars_dtype = pl.Int64
    else:
        polars_dtype = pl.Utf8

    train_labels_df = update_labels_df(
        master_df=train_labels_df,
        new_labels=cur_labels.train_labels,
        output_name=output_name,
    )

    train_labels_df = train_labels_df.with_columns(
        [pl.col(col_name).cast(polars_dtype)]
    )

    valid_labels_df = update_labels_df(
        master_df=valid_labels_df,
        new_labels=cur_labels.valid_labels,
        output_name=output_name,
    ).with_columns([pl.col(col_name).cast(polars_dtype)])

    return train_labels_df, valid_labels_df


def process_survival_output(
    n_bins: int,
    output_name: str,
    tabular_target_labels_info: Dict[str, Any],
    train_ids: Sequence[str],
    valid_ids: Sequence[str],
    all_ids: set[str],
    label_transformers: Dict[str, Any],
    per_modality_missing_ids: Dict[str, set[str]],
    train_labels_df: pl.DataFrame,
    valid_labels_df: pl.DataFrame,
) -> tuple[pl.DataFrame, pl.DataFrame]:
    tabular_info = tabular_target_labels_info[output_name]

    tabular_info_copy = copy(tabular_info)
    tabular_info_copy.con_columns = []

    cur_labels = set_up_train_and_valid_tabular_data(
        tabular_file_info=tabular_info_copy,
        train_ids=train_ids,
        valid_ids=valid_ids,
        do_transform_labels=True,
    )

    assert len(tabular_info.con_columns) == 1
    event_column = tabular_info.cat_columns[0]
    time_column = tabular_info.con_columns[0]

    df_time = pl.read_csv(
        tabular_info.file_path,
        columns=["ID", time_column],
    ).with_columns([pl.col("ID").cast(pl.Utf8), pl.col(time_column).cast(pl.Float32)])

    df_time_train = df_time.filter(pl.col("ID").is_in(train_ids))
    df_time_valid = df_time.filter(pl.col("ID").is_in(valid_ids))

    df_time_train, cur_labels.train_labels = synchronize_missing_survival_values(
        df_time=df_time_train,
        df_labels=cur_labels.train_labels,
        time_column=time_column,
        event_column=event_column,
    )

    df_time_valid, cur_labels.valid_labels = synchronize_missing_survival_values(
        df_time=df_time_valid,
        df_labels=cur_labels.valid_labels,
        time_column=time_column,
        event_column=event_column,
    )

    dur_input = _streamline_duration_transformer_input(
        df_time_train=df_time_train,
        time_column=time_column,
    )
    dur_transformer = fit_duration_transformer(durations=dur_input, n_bins=n_bins)

    cur_labels.train_labels = cur_labels.train_labels.with_columns(
        [
            transform_durations_with_nans(
                df=df_time_train,
                time_column=time_column,
                transformer=dur_transformer,
            ).alias(time_column)
        ]
    )

    cur_labels.valid_labels = cur_labels.valid_labels.with_columns(
        [
            transform_durations_with_nans(
                df=df_time_valid,
                time_column=time_column,
                transformer=dur_transformer,
            ).alias(time_column)
        ]
    )

    cur_labels.label_transformers[time_column] = dur_transformer
    label_transformers[output_name] = cur_labels.label_transformers

    all_labels = cur_labels.all_labels
    cur_ids = set(all_labels.get_column("ID").to_list())
    missing_ids = all_ids.difference(cur_ids)
    per_modality_missing_ids[output_name] = missing_ids

    train_labels_df = update_labels_df(
        master_df=train_labels_df,
        new_labels=cur_labels.train_labels,
        output_name=output_name,
    )
    valid_labels_df = update_labels_df(
        master_df=valid_labels_df,
        new_labels=cur_labels.valid_labels,
        output_name=output_name,
    )

    return train_labels_df, valid_labels_df


def synchronize_missing_survival_values(
    df_time: pl.DataFrame,
    df_labels: pl.DataFrame,
    time_column: str,
    event_column: str,
) -> tuple[pl.DataFrame, pl.DataFrame]:
    na_ids_time = set(
        df_time.filter(pl.col(time_column).is_null()).get_column("ID").to_list()
    )
    na_ids_labels = set(
        df_labels.filter(pl.col(event_column).is_null()).get_column("ID").to_list()
    )

    all_na_ids = list(na_ids_time.union(na_ids_labels))

    df_time_sync = df_time.with_columns(
        [
            pl.when(pl.col("ID").is_in(all_na_ids))
            .then(None)
            .otherwise(pl.col(time_column))
            .alias(time_column)
        ]
    )

    df_labels_sync = df_labels.with_columns(
        [
            pl.when(pl.col("ID").is_in(all_na_ids))
            .then(None)
            .otherwise(pl.col(event_column))
            .alias(event_column)
        ]
    )

    return df_time_sync, df_labels_sync


def _streamline_duration_transformer_input(
    df_time_train: pl.DataFrame,
    time_column: str,
) -> np.ndarray:
    values = (
        df_time_train.filter(pl.col(time_column).is_not_null())
        .get_column(time_column)
        .to_numpy()
    )

    return values.reshape(-1, 1)


def fit_duration_transformer(
    durations: np.ndarray,
    n_bins: int,
) -> KBinsDiscretizer | IdentityTransformer:

    if not n_bins:
        return IdentityTransformer()

    transformer = KBinsDiscretizer(
        n_bins=n_bins,
        encode="ordinal",
        strategy="uniform",
    )

    transformer.fit(durations)

    return transformer


def transform_durations_with_nans(
    df: pl.DataFrame,
    time_column: str,
    transformer: KBinsDiscretizer | IdentityTransformer,
) -> pl.Series:
    values = df.get_column(time_column).to_numpy()
    nan_mask = np.isnan(values)

    if not nan_mask.any():
        return pl.Series(transformer.transform(values.reshape(-1, 1)).flatten())

    non_null_values = values[~nan_mask]
    transformed = transformer.transform(non_null_values.reshape(-1, 1)).flatten()

    result = np.full(len(df), np.nan)
    result[~nan_mask] = transformed

    return pl.Series(result)


def set_up_delayed_target_labels(
    train_ids: Sequence[str],
    valid_ids: Sequence[str],
    output_name: str,
) -> Labels:
    df_train = pl.DataFrame(
        {
            "ID": list(train_ids),
            f"{output_name}": ["DELAYED"] * len(train_ids),
        },
        schema_overrides={f"{output_name}": pl.Categorical},
    )

    df_valid = pl.DataFrame(
        {
            "ID": list(valid_ids),
            f"{output_name}": ["DELAYED"] * len(valid_ids),
        },
        schema_overrides={f"{output_name}": pl.Categorical},
    )

    return Labels(
        train_labels=df_train,
        valid_labels=df_valid,
        label_transformers={},
    )


def find_sequence_output_missing_ids(
    train_ids: Sequence[str],
    valid_ids: Sequence[str],
    output_source: str,
) -> set[str]:

    seq_ids = set(gather_ids_from_data_source(data_source=Path(output_source)))

    train_ids_set = set(train_ids)
    valid_ids_set = set(valid_ids)
    all_ids = train_ids_set.union(valid_ids_set)

    missing_ids = all_ids.difference(seq_ids)

    return missing_ids


def set_up_file_target_labels(
    train_ids: Sequence[str],
    valid_ids: Sequence[str],
    output_config: schemas.OutputConfig,
) -> Labels:
    """
    Note we have the None here because we want to be able to
    handle if we have partially missing output modalities, e.g. in the case
    of output arrays, but we don't want to just drop those IDs.
    """
    output_name = output_config.output_info.output_name
    output_source = output_config.output_info.output_source
    output_name_inner_key = output_config.output_info.output_inner_key

    id_to_data_pointer_mapping = gather_data_pointers_from_data_source(
        data_source=Path(output_source),
        validate=True,
        output_inner_key=output_name_inner_key,
    )

    train_values = [id_to_data_pointer_mapping.get(id_, None) for id_ in train_ids]
    df_train = pl.DataFrame(
        {
            "ID": list(train_ids),
            f"{output_name}": train_values,
        }
    )

    valid_values = [id_to_data_pointer_mapping.get(id_, None) for id_ in valid_ids]
    df_valid = pl.DataFrame(
        {
            "ID": list(valid_ids),
            f"{output_name}": valid_values,
        }
    )

    df_train = df_train.with_columns([pl.col(f"{output_name}").cast(pl.Utf8)])
    df_valid = df_valid.with_columns([pl.col(f"{output_name}").cast(pl.Utf8)])

    return Labels(
        train_labels=df_train,
        valid_labels=df_valid,
        label_transformers={},
    )


def gather_torch_null_missing_ids(labels: pl.DataFrame, output_name: str) -> set[str]:

    missing_ids = (
        labels.filter((pl.col(output_name).is_null())).get_column("ID").to_list()
    )

    return set(str(id_) for id_ in missing_ids)


def gather_data_pointers_from_data_source(
    data_source: Path,
    validate: bool = True,
    output_inner_key: Optional[str] = None,
) -> dict[str, str | int]:
    """
    Disk: ID -> file path
    Deeplake: ID -> integer index
    """
    iterator: (
        Generator[tuple[str, str], None, None] | Generator[tuple[str, int], None, None]
    )
    if is_deeplake_dataset(data_source=str(data_source)):
        assert output_inner_key is not None
        iterator = build_deeplake_available_pointer_iterator(
            data_source=data_source,
            inner_key=output_inner_key,
        )
    else:
        iterator_base = get_file_path_iterator(
            data_source=data_source,
            validate=validate,
        )
        iterator = ((f.stem, str(f)) for f in iterator_base)

    logger.debug("Gathering data pointers from %s.", data_source)
    id_to_pointer_mapping = {}
    for id_, pointer in tqdm(iterator, desc="Progress"):
        if id_ in id_to_pointer_mapping:
            raise ValueError(f"Duplicate ID: {id_}")

        id_to_pointer_mapping[id_] = pointer

    return id_to_pointer_mapping


def build_deeplake_available_pointer_iterator(
    data_source: Path, inner_key: str
) -> Generator[tuple[str, int], None, None]:
    deeplake_ds = load_deeplake_dataset(data_source=str(data_source))
    columns = {col.name for col in deeplake_ds.schema.columns}
    existence_col = f"{inner_key}_exists"
    for int_pointer, row in enumerate(deeplake_ds):
        if is_deeplake_sample_missing(
            row=row,
            existence_col=existence_col,
            columns=columns,
        ):
            pass

        id_ = row["ID"]

        yield id_, int(int_pointer)


def gather_all_ids_from_output_configs(
    output_configs: Sequence[schemas.OutputConfig],
) -> Tuple[str, ...]:
    all_ids: set[str] = set()
    for config in output_configs:
        cur_source = Path(config.output_info.output_source)
        logger.debug("Gathering IDs from %s.", cur_source)
        if cur_source.suffix == ".csv":
            cur_ids = gather_ids_from_tabular_file(file_path=cur_source)
        elif cur_source.is_dir():
            cur_ids = gather_ids_from_data_source(data_source=cur_source)
        else:
            raise NotImplementedError(
                "Only csv and directory data sources are supported."
                f" Got: {cur_source}"
            )
        all_ids.update(cur_ids)

    return tuple(all_ids)


def read_manual_ids_if_exist(
    manual_valid_ids_file: Union[None, str]
) -> Union[Sequence[str], None]:
    if not manual_valid_ids_file:
        return None

    with open(manual_valid_ids_file, "r") as infile:
        manual_ids = tuple(line.strip() for line in infile)

    return manual_ids


def get_tabular_target_file_infos(
    output_configs: Iterable[schemas.OutputConfig],
) -> Dict[str, TabularFileInfo]:

    tabular_files_info = {}

    for output_config in output_configs:
        output_type = output_config.output_info.output_type
        output_name = output_config.output_info.output_name
        if output_type == "tabular":
            output_type_info = output_config.output_type_info
            assert isinstance(output_type_info, TabularOutputTypeConfig)

            tabular_info = TabularFileInfo(
                file_path=Path(output_config.output_info.output_source),
                con_columns=output_type_info.target_con_columns,
                cat_columns=output_type_info.target_cat_columns,
                parsing_chunk_size=output_type_info.label_parsing_chunk_size,
            )
            tabular_files_info[output_name] = tabular_info
        elif output_type == "survival":
            output_type_info = output_config.output_type_info
            assert isinstance(output_type_info, SurvivalOutputTypeConfig)

            tabular_info = TabularFileInfo(
                file_path=Path(output_config.output_info.output_source),
                cat_columns=[output_type_info.event_column],
                con_columns=[output_type_info.time_column],
                parsing_chunk_size=output_type_info.label_parsing_chunk_size,
            )
            tabular_files_info[output_name] = tabular_info

        else:
            continue

    return tabular_files_info



from typing import Callable

from torch import Tensor, nn


class Residual(nn.Module):
    def __init__(self, func: Callable[[Tensor], Tensor]):
        super().__init__()
        self.fn = func

    def forward(self, x, *args, **kwargs):
        return self.fn(x, *args, **kwargs) + x



import typing
from collections import OrderedDict
from dataclasses import dataclass
from typing import Dict, Hashable, Literal, Sequence

from eir.setup import schemas
from eir.setup.input_setup_modules.common import get_default_sequence_specials


@dataclass
class ComputedBytesInputInfo:
    input_config: schemas.InputConfig
    vocab: OrderedDict
    computed_max_length: int


def set_up_bytes_input_for_training(
    input_config: schemas.InputConfig, add_specials: bool = True, *args, **kwargs
) -> ComputedBytesInputInfo:
    specials: list[str] = []
    if add_specials:
        specials = get_default_sequence_specials()

    input_type_info = input_config.input_type_info
    assert isinstance(input_type_info, schemas.ByteInputDataConfig)
    bytes_vocab = build_bytes_vocab(
        byte_encoding=input_type_info.byte_encoding,
        specials=specials,
    )

    if not isinstance(input_type_info.max_length, int):
        raise ValueError(
            "Max length for bytes input only supports raw int values currently."
        )

    bytes_input_info = ComputedBytesInputInfo(
        input_config=input_config,
        vocab=bytes_vocab,
        computed_max_length=input_type_info.max_length,
    )

    return bytes_input_info


def build_bytes_vocab(
    byte_encoding: Literal["uint8"], specials: Sequence[Hashable] = tuple()
) -> typing.OrderedDict[int | Hashable, int]:
    bytes_vocab: typing.OrderedDict[int | Hashable, int] = OrderedDict()

    encoding_to_num_tokens_map = _get_encoding_to_num_tokens_map()
    num_tokens = encoding_to_num_tokens_map[byte_encoding]

    for token in range(num_tokens):
        bytes_vocab[token] = token

    base_num_tokens = len(bytes_vocab)
    for idx, special in enumerate(specials):
        bytes_vocab[special] = base_num_tokens + idx

    return bytes_vocab


def _get_encoding_to_num_tokens_map() -> Dict[str, int]:
    mapping = {"uint8": 256}
    return mapping



from dataclasses import dataclass
from functools import reduce
from typing import Any, Dict, Literal, Protocol, Type

from torch import Tensor, nn

from eir.models.input.array.models_cnn import CNNModelConfig


class ImageModelClassGetterFunction(Protocol):
    def __call__(self, model_type: str) -> Type["ImageWrapperModel"]: ...


def get_image_model_class(model_type: str) -> Type["ImageWrapperModel"]:
    if model_type == "image-wrapper-default":
        return ImageWrapperModel
    else:
        raise ValueError()


@dataclass
class ImageModelConfig:
    """
    :param model_type:
         Which type of image model to use.

    :param model_init_config:
          Configuration / arguments used to initialise model.

    :param num_output_features:
          Number of output final output features from image feature extractor, projected
          with a linear layer, which get passed to fusion module. If set to 0,
          the output from the feature extractor is passed directly as is to the
          fusion module.

    :param pretrained_model:
          Specify whether the model type is assumed to be pretrained and from the
          Pytorch Image Models repository.

    :param freeze_pretrained_model:
          Whether to freeze the pretrained model weights.
    """

    model_type: Literal["cnn"] | str
    model_init_config: CNNModelConfig | Dict[str, Any]

    num_output_features: int = 0

    pretrained_model: bool = False
    freeze_pretrained_model: bool = False


class ImageWrapperModel(nn.Module):
    def __init__(
        self,
        feature_extractor: nn.Module,
        model_config: ImageModelConfig,
        estimated_out_shape: tuple[int, ...],
    ) -> None:
        super().__init__()

        self.feature_extractor = feature_extractor
        self.model_config = model_config
        self.output_shape = estimated_out_shape

    @property
    def num_out_features(self):
        if hasattr(self.feature_extractor, "num_out_features"):
            return self.feature_extractor.num_out_features

        if self.model_config.num_output_features > 0:
            return self.model_config.num_output_features

        return reduce(lambda x, y: x * y, self.output_shape)

    def forward(self, input: Tensor) -> Tensor:

        has_forward_features = hasattr(self.feature_extractor, "forward_features")
        has_num_out_features = self.model_config.num_output_features > 0

        if has_forward_features and not has_num_out_features:
            out = self.feature_extractor.forward_features(input)
        else:
            out = self.feature_extractor(input)

        return out



from dataclasses import dataclass
from typing import Literal, Optional, Sequence


@dataclass
class ArrayOutputTypeConfig:
    """
    :param normalization:
        Which type of normalization to apply to the array data. If ``element``, will
        normalize each element in the array independently. If ``channel``, will
        normalize each channel in the array independently.
        For 'channel', assumes PyTorch format where the channel dimension is the
        first dimension.

    :param adaptive_normalization_max_samples:
        If using adaptive normalization (channel / element),
        how many samples to use to compute the normalization parameters.
        If None, will use all samples.

    :param loss:
        Which loss to use for training the model. Either ``mse`` or ``diffusion``.

    :param diffusion_time_steps:
        Number of time steps to use for diffusion loss. Only used if ``loss`` is
        set to ``diffusion``.
    """

    normalization: Optional[Literal["element", "channel"]] = "channel"
    adaptive_normalization_max_samples: Optional[int] = None
    loss: Literal["mse", "diffusion"] = "mse"
    diffusion_time_steps: Optional[int] = 500


@dataclass
class ArrayOutputSamplingConfig:
    """
    :param manual_inputs:
        Manually specified inputs to use for sequence generation. This is useful
        if you want to generate sequences based on a specific input. Depending
        on the input type, different formats are expected:

        - ``sequence``: A string written directly in the ``.yaml`` file.
        - ``omics``: A file path to NumPy array of shape ``(4, n_SNPs)`` on disk.
        - ``image``: An image file path on disk.
        - ``tabular``: A mapping of :literal:`(column key: value)` written directly
          in the ``.yaml`` file.
        - ``array``: A file path to NumPy array on disk.
        - ``bytes``: A file path to a file on disk.


    :param n_eval_inputs:
        The number of inputs automatically sampled from the validation set for
        sequence generation.
    """

    manual_inputs: Sequence[dict[str, str]] = tuple()
    n_eval_inputs: int = 10



from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import torch
from lifelines import KaplanMeierFitter
from matplotlib import pyplot as plt

from eir.experiment_io.output_object_io import get_output_serialization_path
from eir.setup.output_setup_modules.survival_output_setup import (
    ComputedSurvivalOutputInfo,
)
from eir.train_utils import utils
from eir.train_utils.metrics import (
    al_step_metric_dict,
    filter_survival_missing_targets,
    general_torch_to_numpy,
)

if TYPE_CHECKING:
    from eir.train import Experiment


def save_survival_evaluation_results_wrapper(
    val_outputs: dict[str, dict[str, torch.Tensor]],
    val_labels: dict[str, dict[str, torch.Tensor]],
    val_ids: dict[str, dict[str, list[str]]],
    iteration: int,
    experiment: "Experiment",
    evaluation_metrics: al_step_metric_dict,
) -> None:
    for output_name, output_object in experiment.outputs.items():
        output_type = output_object.output_config.output_info.output_type
        if output_type != "survival":
            continue

        assert isinstance(output_object, ComputedSurvivalOutputInfo)

        output_type_info = output_object.output_config.output_type_info
        time_name = output_type_info.time_column
        event_name = output_type_info.event_column
        model_type = (
            "cox" if output_type_info.loss_function == "CoxPHLoss" else "discrete"
        )

        cur_sample_output_folder = utils.prepare_sample_output_folder(
            output_folder=experiment.configs.gc.be.output_folder,
            column_name=event_name,
            output_name=output_name,
            iteration=iteration,
        )

        ids = val_ids[output_name][event_name]
        model_outputs = val_outputs[output_name][event_name]
        events = val_labels[output_name][event_name]
        times = val_labels[output_name][time_name]

        filtered = filter_survival_missing_targets(
            model_outputs=model_outputs,
            events=events,
            times=times,
            cur_ids=ids,
        )

        model_outputs = filtered.model_outputs
        events = general_torch_to_numpy(tensor=filtered.events)
        events = events.astype(int)
        times = general_torch_to_numpy(tensor=filtered.times)
        ids = filtered.ids

        if model_type == "discrete":
            transformers = output_object.target_transformers
            time_kbins_transformer = transformers[time_name]
            time_bins = time_kbins_transformer.bin_edges_[0]
            times_binned = times
            it_func = time_kbins_transformer.inverse_transform
            times = it_func(times_binned.reshape(-1, 1)).flatten()

            hazards = torch.sigmoid(model_outputs).numpy()
            survival_probs = np.cumprod(1 - hazards, 1)
            time_bins_except_last = time_bins[:-1]

            plot_discrete_survival_curves(
                times=times,
                events=events,
                predicted_probs=survival_probs,
                time_bins=time_bins_except_last,
                output_folder=cur_sample_output_folder,
            )

            df = pd.DataFrame(
                {
                    "ID": ids,
                    time_name: times,
                    event_name: events,
                    "Predicted_Risk": hazards[:, -1],
                }
            )

            for i, t in enumerate(time_bins_except_last):
                df[f"Surv_Prob_t{i}"] = survival_probs[:, i]

            plot_discrete_individual_survival_curves(
                df=df,
                time_bins=time_bins_except_last,
                output_folder=str(cur_sample_output_folder),
                n_samples=5,
            )

        else:
            risk_scores = model_outputs.numpy()

            unique_times, baseline_hazard = estimate_baseline_hazard(
                times=times,
                events=events,
                risk_scores=risk_scores,
            )
            baseline_survival = np.exp(-np.cumsum(baseline_hazard))

            run_folder = Path(experiment.configs.gc.be.output_folder)
            cur_serialization_path = get_output_serialization_path(
                output_name=output_name,
                output_type=output_type,
                run_folder=Path(experiment.configs.gc.be.output_folder),
            )

            em = evaluation_metrics
            cur_performance_avg = em["average"]["average"]["perf-average"]

            maybe_save_survival_hazards_and_times(
                serialization_output_folder=cur_serialization_path,
                run_folder=run_folder,
                cur_performance=cur_performance_avg,
                baseline_hazards=baseline_hazard,
                unique_times=unique_times,
            )

            max_time = np.max(times)
            time_points = np.linspace(0, max_time, 100)

            survival_probs = calculate_cox_survival_probs(
                risk_scores=risk_scores,
                unique_times=unique_times,
                baseline_survival=baseline_survival,
                time_points=time_points,
            )

            plot_cox_survival_curves(
                times=times,
                events=events,
                risk_scores=risk_scores,
                unique_times=unique_times,
                baseline_survival=baseline_survival,
                time_points=time_points,
                output_folder=cur_sample_output_folder,
            )

            base_data = {
                "ID": ids,
                time_name: times,
                event_name: events,
                "Risk_Score": risk_scores.squeeze(),
            }

            surv_prob_cols = {
                f"Surv_Prob_t{i}": survival_probs[:, i]
                for i in range(survival_probs.shape[1])
            }

            df = pd.DataFrame({**base_data, **surv_prob_cols})

            plot_cox_individual_curves(
                df=df,
                time_points=time_points,
                output_folder=str(cur_sample_output_folder),
                n_samples=5,
            )

        csv_path = f"{cur_sample_output_folder}/survival_predictions.csv"
        df.to_csv(csv_path, index=False)


def maybe_save_survival_hazards_and_times(
    serialization_output_folder: Path,
    run_folder: Path,
    cur_performance: float,
    baseline_hazards: np.ndarray,
    unique_times: np.ndarray,
) -> None:

    sof = serialization_output_folder
    baseline_hazard_path = sof / "baseline_hazard.npy"
    baseline_unique_times_path = sof / "baseline_unique_times.npy"

    if not baseline_hazard_path.exists():
        np.save(baseline_hazard_path, baseline_hazards)
        np.save(baseline_unique_times_path, unique_times)
    else:
        is_current_iter_best = check_is_current_iter_best(
            run_folder=run_folder,
            cur_performance=cur_performance,
        )

        if is_current_iter_best:
            np.save(baseline_hazard_path, baseline_hazards)
            np.save(baseline_unique_times_path, unique_times)


def check_is_current_iter_best(
    run_folder: Path,
    cur_performance: float,
) -> bool:
    validation_history_file = run_folder / "validation_average_history.log"
    df = pd.read_csv(validation_history_file)

    best_performance = df["perf-average"].max()
    return cur_performance >= best_performance


def estimate_baseline_hazard(
    times: np.ndarray,
    events: np.ndarray,
    risk_scores: np.ndarray,
) -> tuple[np.ndarray, np.ndarray]:
    sort_idx = np.argsort(times)
    times = times[sort_idx]
    events = events[sort_idx]
    risk_scores = risk_scores[sort_idx]

    unique_times = np.unique(times[events == 1])
    baseline_hazard = np.zeros_like(unique_times)

    risk_scores_exp = np.exp(risk_scores)

    for i, t in enumerate(unique_times):
        events_at_t = events[times == t]
        n_events = np.sum(events_at_t)

        at_risk = times >= t
        risk_set_sum = np.sum(risk_scores_exp[at_risk])

        if risk_set_sum > 0:
            baseline_hazard[i] = n_events / risk_set_sum

    return unique_times, baseline_hazard


def calculate_cox_survival_probs(
    risk_scores: np.ndarray,
    unique_times: np.ndarray,
    baseline_survival: np.ndarray,
    time_points: np.ndarray,
) -> np.ndarray:
    interpolated_baseline = np.interp(
        time_points, unique_times, baseline_survival, right=baseline_survival[-1]
    )

    survival_probs = np.zeros((len(risk_scores), len(time_points)))
    for i, risk_score in enumerate(risk_scores):
        survival_probs[i] = interpolated_baseline ** np.exp(risk_score)

    return survival_probs


def plot_cox_survival_curves(
    times: np.ndarray,
    events: np.ndarray,
    risk_scores: np.ndarray,
    unique_times: np.ndarray,
    baseline_survival: np.ndarray,
    time_points: np.ndarray,
    output_folder: Path,
) -> None:
    kmf = KaplanMeierFitter()
    kmf.fit(times, events)

    plt.figure(figsize=(10, 6))
    kmf.plot_survival_function(label="Kaplan-Meier Estimate")

    risk_quantiles = np.percentile(risk_scores, [25, 50, 75])

    for q, risk in zip(
        ["Low Risk (25%)", "Median Risk", "High Risk (75%)"], risk_quantiles
    ):
        interp_baseline = np.interp(time_points, unique_times, baseline_survival)
        surv = interp_baseline ** np.exp(risk)
        plt.plot(time_points, surv, label=f"{q}")

    plt.xlabel("Time")
    plt.ylabel("Survival Probability")
    plt.legend()
    plt.title("Kaplan-Meier vs Model Predictions")
    plt.grid(True, alpha=0.3)
    plt.savefig(f"{output_folder}/cox_survival_curves.pdf")
    plt.close()


def plot_cox_individual_curves(
    df: pd.DataFrame,
    time_points: np.ndarray,
    output_folder: str,
    n_samples: int = 5,
) -> None:
    random_samples = df.sample(n=n_samples)

    plt.figure(figsize=(12, 8))

    for _, row in random_samples.iterrows():
        surv_probs = row[
            [col for col in df.columns if col.startswith("Surv_Prob_t")]
        ].to_numpy()
        plt.plot(
            time_points,
            surv_probs,
            label=f"ID: {row['ID']} (Risk Score: {row['Risk_Score']:.2f})",
        )

    plt.xlabel("Time")
    plt.ylabel("Survival Probability")
    plt.title(f"Individual Survival Curves for {n_samples} Random Samples")
    plt.legend()
    plt.grid(True, alpha=0.3)
    plt.savefig(f"{output_folder}/cox_individual_curves.pdf")
    plt.close()


def plot_discrete_survival_curves(
    times,
    events,
    predicted_probs,
    time_bins,
    output_folder,
):
    kmf = KaplanMeierFitter()
    kmf.fit(times, events)

    plt.figure(figsize=(10, 6))
    kmf.plot_survival_function(label="Kaplan-Meier Estimate")

    mean_predicted = predicted_probs.mean(axis=0)
    plt.step(time_bins, mean_predicted, where="post", label="Mean Predicted")

    plt.xlabel("Time")
    plt.ylabel("Survival Probability")
    plt.legend()
    plt.title("Kaplan-Meier vs Mean Predicted Survival")
    plt.savefig(f"{output_folder}/survival_curves.png")
    plt.close()


def plot_discrete_individual_survival_curves(
    df: pd.DataFrame,
    time_bins: np.ndarray,
    output_folder: str,
    n_samples: int = 5,
) -> None:
    random_samples = df.sample(n=n_samples)

    plt.figure(figsize=(12, 8))

    for _, row in random_samples.iterrows():
        surv_probs = row[
            [col for col in df.columns if col.startswith("Surv_Prob_t")]
        ].to_numpy()

        plt.step(x=time_bins, y=surv_probs, where="post", label=f"ID: {row['ID']}")

    plt.xlabel("Time")
    plt.ylabel("Survival Probability")
    plt.title(f"Survival Curves for {n_samples} Random Samples")
    plt.legend()
    plt.grid(True, alpha=0.3)

    plt.savefig(f"{output_folder}/individual_survival_curves.pdf")
    plt.close()



import reprlib
from dataclasses import dataclass
from functools import lru_cache
from pathlib import Path
from typing import (
    Any,
    Dict,
    Generator,
    List,
    Literal,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    Type,
    Union,
)

import numpy as np
import polars as pl

pl.enable_string_cache()
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import KBinsDiscretizer, LabelEncoder, StandardScaler
from tqdm import tqdm

from eir.data_load.data_source_modules.deeplake_ops import (
    is_deeplake_dataset,
    is_deeplake_sample_missing,
    load_deeplake_dataset,
)
from eir.setup.schemas import InputConfig
from eir.target_setup.target_setup_utils import IdentityTransformer
from eir.train_utils.utils import get_seed
from eir.utils.logging import get_logger

logger = get_logger(name=__name__, tqdm_compatible=True)

# Type Aliases
al_train_val_dfs = tuple[pl.DataFrame, pl.DataFrame]

# e.g. 'Asia' or '5' for categorical or 1.511 for continuous
al_label_values_raw = float | int | Path
al_sample_labels_raw = dict[str, al_label_values_raw]
al_label_dict = dict[str, al_sample_labels_raw]
al_target_labels = pl.DataFrame
al_target_columns = dict[Literal["con", "cat"], list[str]]
al_label_transformers_object = Union[
    StandardScaler,
    LabelEncoder,
    KBinsDiscretizer,
    IdentityTransformer,
]
al_label_transformers = dict[str, al_label_transformers_object]


@dataclass
class Labels:
    train_labels: pl.DataFrame
    valid_labels: pl.DataFrame
    label_transformers: al_label_transformers

    @property
    def all_labels(self) -> pl.DataFrame:
        return pl.concat([self.train_labels, self.valid_labels])


@dataclass
class TabularFileInfo:
    file_path: Path
    con_columns: Sequence[str]
    cat_columns: Sequence[str]
    parsing_chunk_size: Union[None, int] = None


def set_up_train_and_valid_tabular_data(
    tabular_file_info: TabularFileInfo,
    train_ids: Sequence[str],
    valid_ids: Sequence[str],
    impute_missing: bool = False,
    do_transform_labels: bool = True,
) -> Labels:
    """
    Splits and does split based processing (e.g. scaling validation set with training
    set for regression) on the labels.
    """
    if len(tabular_file_info.con_columns) + len(tabular_file_info.cat_columns) < 1:
        raise ValueError(f"No label columns specified in {tabular_file_info}.")

    parse_wrapper = get_label_parsing_wrapper(
        label_parsing_chunk_size=tabular_file_info.parsing_chunk_size
    )
    ids_to_keep = list(train_ids) + list(valid_ids)
    df_labels = parse_wrapper(
        label_file_tabular_info=tabular_file_info,
        ids_to_keep=ids_to_keep,
    )
    _validate_df(df=df_labels)

    df_labels_train, df_labels_valid = _split_df_by_ids(
        df=df_labels,
        train_ids=list(train_ids),
        valid_ids=list(valid_ids),
    )
    pre_check_label_df(df=df_labels_train, name="Training DataFrame")
    pre_check_label_df(df=df_labels_valid, name="Validation DataFrame")
    check_train_valid_df_sync(
        df_train=df_labels_train,
        df_valid=df_labels_valid,
        cat_columns=tabular_file_info.cat_columns,
    )

    label_transformers: al_label_transformers = {}
    if do_transform_labels:
        df_labels_train, df_labels_valid, label_transformers = (
            _process_train_and_label_dfs(
                tabular_info=tabular_file_info,
                df_labels_train=df_labels_train,
                df_labels_valid=df_labels_valid,
                impute_missing=impute_missing,
            )
        )

    labels_data_object = Labels(
        train_labels=df_labels_train,
        valid_labels=df_labels_valid,
        label_transformers=label_transformers,
    )

    return labels_data_object


def _get_fit_label_transformers(
    df_labels_train: pl.DataFrame,
    df_labels_full: pl.DataFrame,
    label_columns: al_target_columns,
    impute_missing: bool,
) -> al_label_transformers:
    label_transformers = {}

    for column_type in label_columns:
        label_columns_for_cur_type = label_columns[column_type]

        if label_columns_for_cur_type:
            logger.debug(
                "Fitting transformers on %s label columns %s",
                column_type,
                label_columns_for_cur_type,
            )

        for label_column in label_columns_for_cur_type:
            if column_type == "con":
                cur_values = df_labels_train.get_column(name=label_column)
            elif column_type == "cat":
                cur_values = df_labels_full.get_column(name=label_column)
            else:
                raise ValueError(f"Unknown column type: {column_type}")

            cur_transformer = _get_transformer(column_type=column_type)
            cur_target_transformer_fit = _fit_transformer_on_label_column(
                column_series=cur_values,
                transformer=cur_transformer,
                impute_missing=impute_missing,
            )
            label_transformers[label_column] = cur_target_transformer_fit

    return label_transformers


def _get_transformer(column_type):
    if column_type in ("con", "extra_con"):
        return StandardScaler()
    elif column_type == "cat":
        return LabelEncoder()

    raise ValueError()


def _fit_transformer_on_label_column(
    column_series: pl.Series,
    transformer: al_label_transformers_object,
    impute_missing: bool,
) -> al_label_transformers_object:
    if column_series.dtype == pl.Categorical or column_series.dtype == pl.Utf8:
        unique_values = column_series.unique().drop_nulls()
        series_values = unique_values.to_numpy()
        if impute_missing:
            series_values = np.append(series_values, "__NULL__")
    else:
        series_values = column_series.drop_nulls().to_numpy()

    series_values_streamlined = streamline_values_for_transformers(
        transformer=transformer,
        values=series_values,
    )

    transformer.fit(series_values_streamlined)
    return transformer


def streamline_values_for_transformers(
    transformer: al_label_transformers_object, values: np.ndarray
) -> np.ndarray:
    """
    LabelEncoder() expects a 1D array, whereas StandardScaler() expects a 2D one.
    """

    if isinstance(transformer, (StandardScaler, KBinsDiscretizer, IdentityTransformer)):
        values_reshaped = values.reshape(-1, 1)
        return values_reshaped

    return values


def transform_label_df(
    df_labels: pl.DataFrame,
    label_transformers: al_label_transformers,
    impute_missing: bool,
) -> pl.DataFrame:
    expr = []

    for column_name, transformer_instance in label_transformers.items():
        series_values = df_labels.get_column(column_name).to_numpy()
        transform_func = transformer_instance.transform

        if impute_missing:
            series_values_streamlined = streamline_values_for_transformers(
                transformer=transformer_instance,
                values=series_values,
            )
            transformed = transform_func(series_values_streamlined)
            expr.append(pl.lit(transformed.squeeze()).alias(column_name))
        else:
            match transformer_instance:
                case StandardScaler():
                    non_nan_mask = ~df_labels.get_column(column_name).is_null()
                case LabelEncoder():
                    non_nan_mask = ~df_labels.get_column(column_name).is_null()
                case KBinsDiscretizer() | IdentityTransformer():
                    non_nan_mask = ~df_labels.get_column(column_name).is_null()
                case _:
                    raise ValueError(
                        f"Unknown transformer type: {type(transformer_instance)}"
                    )

            non_nan_values = series_values[non_nan_mask.to_numpy()]
            series_values_streamlined = streamline_values_for_transformers(
                transformer=transformer_instance,
                values=non_nan_values,
            )

            transformed_values = transform_func(series_values_streamlined).squeeze()

            result = np.full(len(series_values), np.nan, dtype=np.float32)
            result[non_nan_mask.to_numpy()] = transformed_values

            if isinstance(transformer_instance, LabelEncoder):
                result[~non_nan_mask.to_numpy()] = np.nan

            expr.append(pl.lit(result).alias(column_name))

    df_final = df_labels.with_columns(expr)
    return df_final


class LabelDFParseWrapperProtocol(Protocol):
    def __call__(
        self,
        label_file_tabular_info: TabularFileInfo,
        ids_to_keep: Union[None, Sequence[str]],
    ) -> pl.DataFrame: ...


def get_label_parsing_wrapper(
    label_parsing_chunk_size: Union[None, int]
) -> LabelDFParseWrapperProtocol:
    if label_parsing_chunk_size is None:
        return label_df_parse_wrapper
    return chunked_label_df_parse_wrapper


def _validate_df(df: pl.DataFrame) -> None:
    duplicate_counts = (
        df.group_by("ID").agg(pl.count().alias("count")).filter(pl.col("count") > 1)
    )

    if duplicate_counts.height > 0:
        duplicated_ids = duplicate_counts.select("ID").limit(10).to_series().to_list()
        duplicated_indices_str = ", ".join(map(str, duplicated_ids))
        raise ValueError(
            f"Found duplicated indices in the dataframe. "
            f"Random examples: {duplicated_indices_str}. "
            f"Please make sure that the indices in the ID column are unique."
        )


def label_df_parse_wrapper(
    label_file_tabular_info: TabularFileInfo,
    ids_to_keep: Union[None, Sequence[str]],
) -> pl.DataFrame:

    all_label_columns, dtypes = _get_all_label_columns_and_dtypes(
        cat_columns=label_file_tabular_info.cat_columns,
        con_columns=label_file_tabular_info.con_columns,
    )

    df_labels = _load_label_df(
        label_fpath=label_file_tabular_info.file_path,
        columns=all_label_columns,
        dtypes=dtypes,
    )

    df_labels_filtered = _filter_ids_from_label_df(
        df_labels=df_labels,
        ids_to_keep=ids_to_keep,
    )

    supplied_columns = get_passed_in_columns(tabular_info=label_file_tabular_info)
    df_column_filtered = _drop_not_needed_label_columns(
        df=df_labels_filtered,
        needed_label_columns=supplied_columns,
    )

    df_cat_str = ensure_categorical_columns_and_format(df=df_column_filtered)

    df_final = _check_parsed_label_df(
        df_labels=df_cat_str,
        supplied_label_columns=supplied_columns,
    )

    return df_final


def chunked_label_df_parse_wrapper(
    label_file_tabular_info: TabularFileInfo,
    ids_to_keep: Union[None, Sequence[str]],
) -> pl.DataFrame:
    label_columns, dtypes = _get_all_label_columns_and_dtypes(
        cat_columns=label_file_tabular_info.cat_columns,
        con_columns=label_file_tabular_info.con_columns,
    )

    dtypes = _ensure_id_str_dtype(dtypes=dtypes)

    assert isinstance(label_file_tabular_info.parsing_chunk_size, int)

    stream = pl.read_csv_batched(
        source=label_file_tabular_info.file_path,
        schema_overrides=dtypes,
        has_header=True,
        batch_size=label_file_tabular_info.parsing_chunk_size,
    )

    supplied_columns = get_passed_in_columns(tabular_info=label_file_tabular_info)

    def process_chunk(chunk: pl.DataFrame) -> Optional[pl.DataFrame]:
        if chunk.height == 0:
            return None

        chunk = chunk.select([pl.col("ID").cast(pl.Utf8), pl.all().exclude("ID")])

        df_labels_filtered = _filter_ids_from_label_df(
            df_labels=chunk,
            ids_to_keep=ids_to_keep,
        )

        if df_labels_filtered.height == 0:
            return None

        return _drop_not_needed_label_columns(
            df=df_labels_filtered,
            needed_label_columns=supplied_columns,
        )

    processed_chunks: list[pl.DataFrame] = []

    while True:
        next_batches = stream.next_batches(1)

        if not next_batches:
            break

        for chunk in next_batches:
            processed_chunk = process_chunk(chunk=chunk)
            if processed_chunk is not None:
                pre_check_label_df(
                    df=processed_chunk,
                    name=f"{label_file_tabular_info.file_path}:chunk",
                )
                processed_chunks.append(processed_chunk)

    if not processed_chunks:
        return pl.DataFrame(schema=dtypes)

    df_concat = pl.concat(processed_chunks)

    df_cat_str = ensure_categorical_columns_and_format(df=df_concat)

    df_final = _check_parsed_label_df(
        df_labels=df_cat_str,
        supplied_label_columns=supplied_columns,
    )

    return df_final


def ensure_categorical_columns_and_format(df: pl.DataFrame) -> pl.DataFrame:
    expr = []
    for column in df.columns:
        if column == "ID":
            continue

        col_dtype = df.schema[column]
        if col_dtype in [pl.Object, pl.Categorical, pl.Utf8]:
            expr.append(pl.col(column).cast(pl.Categorical).alias(column))
        else:
            expr.append(pl.col(column))

    return df.with_columns(expr)


def gather_all_ids_from_all_inputs(
    input_configs: Sequence[InputConfig],
) -> Tuple[str, ...]:
    ids = set()
    for input_config in input_configs:
        cur_source = Path(input_config.input_info.input_source)
        cur_type = input_config.input_info.input_type
        if cur_type in ["omics", "sequence", "bytes", "image", "array"]:
            cur_ids = gather_ids_from_data_source(data_source=cur_source)

        elif cur_type == "tabular":
            cur_ids = gather_ids_from_tabular_file(
                file_path=Path(input_config.input_info.input_source)
            )
        else:
            raise NotImplementedError(
                f"ID gather not implemented for type {cur_type} (source: {cur_source})"
            )

        cur_ids_set = set(cur_ids)
        ids.update(cur_ids_set)

    return tuple(ids)


def gather_ids_from_data_source(
    data_source: Path,
    validate: bool = True,
) -> Tuple[str, ...]:
    iterator: Generator[str, None, None] | Generator[Path, None, None]
    if is_deeplake_dataset(data_source=str(data_source)):
        iterator = build_deeplake_available_id_iterator(
            data_source=data_source,
            inner_key="ID",
        )
    elif data_source.suffix == ".csv":
        ids = gather_ids_from_tabular_file(file_path=data_source)
        iterator = (str(i) for i in ids)
    else:
        iterator = get_file_path_iterator(data_source=data_source, validate=validate)
        iterator = (i.stem for i in iterator)

    logger.debug("Gathering IDs from %s.", data_source)
    all_ids = tuple(i for i in tqdm(iterator, desc="Progress"))

    return all_ids


def build_deeplake_available_id_iterator(
    data_source: Path, inner_key: str
) -> Generator[str, None, None]:
    deeplake_ds = load_deeplake_dataset(data_source=str(data_source))
    columns = {col.name for col in deeplake_ds.schema.columns}
    existence_col = f"{inner_key}_exists"
    for row in deeplake_ds:
        if is_deeplake_sample_missing(
            row=row,
            existence_col=existence_col,
            columns=columns,
        ):
            continue

        id_ = row["ID"]

        yield id_


@lru_cache()
def gather_ids_from_tabular_file(file_path: Path) -> Tuple[str, ...]:
    df = pl.read_csv(file_path, columns=["ID"])
    all_ids = tuple(df.select(pl.col("ID").cast(pl.Utf8)).to_series().to_list())
    return all_ids


def get_file_path_iterator(
    data_source: Path, validate: bool = True
) -> Generator[Path, None, None]:
    def _file_iterator(file_path: Path):
        with open(str(file_path), "r") as infile:
            for line in infile:
                path = Path(line.strip())

                if validate:
                    if not path.exists():
                        raise FileNotFoundError(
                            f"Could not find array {path} listed in {data_source}."
                        )

                yield path

    if data_source.is_dir():
        return data_source.rglob("*")
    elif data_source.is_file():
        return _file_iterator(file_path=data_source)

    if not data_source.exists():
        raise FileNotFoundError("Could not find data source %s.", data_source)
    raise ValueError(
        "Data source %s is neither recognized as a file nor folder.", data_source
    )


def _get_all_label_columns_and_dtypes(
    cat_columns: Sequence[str],
    con_columns: Sequence[str],
) -> Tuple[Sequence[str], dict[str, Type[pl.Categorical] | Type[pl.Float32]]]:
    supplied_label_columns = _get_column_dtypes(
        cat_columns=cat_columns,
        con_columns=con_columns,
    )

    all_cols_and_dtypes = {**supplied_label_columns}
    all_cols = tuple(all_cols_and_dtypes.keys())
    supplied_dtypes = {k: v for k, v in all_cols_and_dtypes.items() if v is not None}

    return all_cols, supplied_dtypes


def _get_column_dtypes(
    cat_columns: Sequence[str], con_columns: Sequence[str]
) -> dict[str, Type[pl.Categorical] | Type[pl.Float32]]:
    dtypes: dict[str, Type[pl.Categorical] | Type[pl.Float32]] = {}

    for cat_column in cat_columns:
        dtypes[cat_column] = pl.Categorical
    for con_column in con_columns:
        dtypes[con_column] = pl.Float32

    return dtypes


def _load_label_df(
    label_fpath: Path,
    columns: Sequence[str],
    dtypes: Optional[dict[str, type[pl.Categorical] | type[pl.Float32]]] = None,
) -> pl.DataFrame:
    dtypes = _ensure_id_str_dtype(dtypes=dtypes)

    logger.debug("Reading in labelfile: %s. ID is read as str dtype.", label_fpath)

    columns_with_id_col = ["ID"] + list(columns)
    available_columns = _get_currently_available_columns(
        label_fpath=label_fpath,
        requested_columns=columns_with_id_col,
    )

    df_labels = pl.read_csv(
        label_fpath,
        columns=available_columns,
        schema_overrides=dtypes,
    )

    df_labels = df_labels.select([pl.col("ID").cast(pl.Utf8), pl.all().exclude("ID")])

    pre_check_label_df(df=df_labels, name=str(label_fpath))

    return df_labels


def pre_check_label_df(df: pl.DataFrame, name: str) -> None:
    for column in df.columns:
        if df.get_column(column).is_null().all():
            raise ValueError(
                f"All values are NULL in column '{column}' from {name}. "
                f"This can either be due to all values in the column actually "
                f"being NULL, or an unfavorable split happened during train/validation "
                f"splitting, causing all values in the training split for the column "
                f"to be NULL. For now this will raise an error, but might be handled "
                f"in the future."
                f" In any case, please remove '{column}' as an input/target for now.",
            )


def check_train_valid_df_sync(
    df_train: pl.DataFrame, df_valid: pl.DataFrame, cat_columns: Sequence[str]
) -> None:
    for col in cat_columns:
        train_values = set(df_train.get_column(col).unique().to_list())
        valid_values = set(df_valid.get_column(col).unique().to_list())

        mismatched_values = valid_values - train_values

        if mismatched_values:
            total_mismatch_count = df_valid.filter(
                pl.col(col).is_in(list(mismatched_values))
            ).height
            total_count = df_valid.height + df_train.height
            percentage = (total_mismatch_count / total_count) * 100

            error_message = (
                f"Mismatched values found in column '{col}': {mismatched_values}. "
                f"Count: {total_mismatch_count}, "
                f"Percentage of total data: {percentage:.2f}%.\n"
                f"This happens as there are values in the validation set that "
                f"are not present in the training set. "
                f"This can happen by chance when working with sparse/rare "
                f"categorical values and/or small datasets. These values will still be "
                f"encoded and used during validation, but might result in nonsensical "
                f"predictions as the model never saw them during training.\n"
                f"One approach to fix this "
                f"is trying a different train/validation split, "
                f"which can be done by:\n"
                f"  1. Running the command with a different seed, "
                f"e.g. EIR_SEED=1 eirtrain ...\n"
                f"  2. Manually specifying the train/validation split, "
                f" using manual_valid_ids_file: <.txt file> in the global "
                f"configuration.\n"
                f"  3. Using a different train/validation split ratio, using "
                f" the valid_size parameter in the global configuration."
                f"Other solutions include:\n"
                f"  4. Skipping the column if it's not crucial.\n"
                f"  5. Binning sparse/rare values into broader categories.\n"
                f"Finally, this will very likely raise an error during testing, "
                f"as the label encoder will likely encounter values "
                f"it has never seen before."
            )
            logger.warning(error_message)


def _ensure_id_str_dtype(dtypes: Union[dict[str, Any], None]) -> dict[str, Any]:
    if dtypes is None:
        dtypes = {"ID": pl.Utf8}
    elif "ID" not in dtypes.keys():
        dtypes["ID"] = pl.Utf8

    return dtypes


def _get_currently_available_columns(
    label_fpath: Path,
    requested_columns: List[str],
) -> List[str]:
    label_file_columns_set = set(pl.read_csv(source=label_fpath, n_rows=0).columns)

    requested_columns_set = set(requested_columns)

    missing_columns = requested_columns_set - label_file_columns_set
    if missing_columns:
        raise ValueError(f"Could not find columns {missing_columns} in {label_fpath}.")

    available_columns = requested_columns_set.intersection(label_file_columns_set)

    return list(available_columns)


def _filter_ids_from_label_df(
    df_labels: pl.DataFrame,
    ids_to_keep: Union[None, Sequence[str]] = None,
) -> pl.DataFrame:
    if not ids_to_keep:
        return df_labels

    no_labels = df_labels.height

    df_filtered = df_labels.filter(pl.col("ID").is_in(ids_to_keep))

    no_dropped = no_labels - df_filtered.height

    logger.debug(
        "Removed %d file IDs from label file based on IDs present in data folder.",
        no_dropped,
    )

    return df_filtered


def _check_parsed_label_df(
    df_labels: pl.DataFrame,
    supplied_label_columns: Sequence[str],
) -> pl.DataFrame:
    """
    Validate DataFrame structure and types using Polars.
    """
    missing_columns = set(supplied_label_columns) - set(df_labels.columns)
    if missing_columns:
        raise ValueError(
            f"Columns asked for in CL args ({missing_columns}) "
            f"missing from columns in label dataframe (with columns "
            f"{df_labels.columns}. The missing columns are not "
            f"found in the raw label file."
        )

    id_dtype = df_labels.schema["ID"]
    assert id_dtype == pl.Utf8, f"ID column must be string type, got {id_dtype}"

    for column in df_labels.columns:
        dtype = df_labels.schema[column]

        if column == "ID":
            assert dtype == pl.Utf8, f"ID column must be string type, got {dtype}"
            continue

        assert dtype in [
            pl.Float32,
            pl.Categorical,
        ], f"Column {column} has invalid type {dtype}"

        if dtype == pl.Categorical:
            non_null_vals = (
                df_labels.filter(pl.col(column).is_not_null())
                .get_column(column)
                .unique()
            )

            if len(non_null_vals) > 0:
                vals_and_types = [(val, type(val)) for val in non_null_vals]
                all_str = all(isinstance(val, str) for val, _ in vals_and_types)

                if not all_str:
                    raise ValueError(
                        f"Non-string values found in string column {column}"
                        f"Got values {vals_and_types}."
                    )

    return df_labels


def get_passed_in_columns(tabular_info: TabularFileInfo) -> Sequence[str]:
    cat_columns = tabular_info.cat_columns
    con_columns = tabular_info.con_columns

    passed_in_columns = list(cat_columns) + list(con_columns)

    return passed_in_columns


def _drop_not_needed_label_columns(
    df: pl.DataFrame, needed_label_columns: Sequence[str]
) -> pl.DataFrame:
    needed_columns = ["ID"] + [col for col in needed_label_columns]
    return df.select(needed_columns)


def _split_df_by_ids(
    df: pl.DataFrame,
    train_ids: list[str],
    valid_ids: list[str],
) -> al_train_val_dfs:
    df_labels_train = df.filter(pl.col("ID").is_in(train_ids))
    df_labels_valid = df.filter(pl.col("ID").is_in(valid_ids))

    msg = (
        "Total number of rows in train and validation "
        "sets doesn't match original DataFrame"
    )
    assert df_labels_train.height + df_labels_valid.height == df.height, msg

    return df_labels_train, df_labels_valid


def split_ids(
    ids: Sequence[str],
    valid_size: Union[int, float],
    manual_valid_ids: Union[None, Sequence[str]] = None,
) -> Tuple[Sequence[str], Sequence[str]]:
    """
    We sort here to ensure that we get the same splits every time.
    """

    if manual_valid_ids:
        logger.info(
            "Doing a manual split into validation set with %d IDs read from file.",
            len(manual_valid_ids),
        )
        train_ids, valid_ids = _split_ids_manual(
            ids=ids, manual_valid_ids=manual_valid_ids
        )

    else:
        seed, _ = get_seed()
        ids_sorted = sorted(list(ids))
        train_ids, valid_ids = train_test_split(
            ids_sorted, test_size=valid_size, random_state=seed
        )

    assert len(train_ids) + len(valid_ids) == len(ids)
    assert set(train_ids).isdisjoint(set(valid_ids))

    return train_ids, valid_ids


def _split_ids_manual(
    ids: Sequence[str], manual_valid_ids: Sequence[str]
) -> Tuple[Sequence[str], Sequence[str]]:
    ids_set = set(ids)
    not_found = tuple(i for i in manual_valid_ids if i not in ids_set)
    if not_found:
        raise ValueError(
            f"Did not find {len(not_found)} manual validation IDs "
            f"'{reprlib.repr(not_found)}' among those IDs. Possibly some validation "
            f"IDs are not present, or this is a bug."
        )

    train_ids_set = set(ids) - set(manual_valid_ids)
    train_ids = list(train_ids_set)
    valid_ids = list(manual_valid_ids)

    return train_ids, valid_ids


def _process_train_and_label_dfs(
    tabular_info: TabularFileInfo,
    df_labels_train: pl.DataFrame,
    df_labels_valid: pl.DataFrame,
    impute_missing: bool,
) -> Tuple[pl.DataFrame, pl.DataFrame, al_label_transformers]:
    train_con_means = _get_con_manual_vals_dict(
        df=df_labels_train,
        con_columns=tabular_info.con_columns,
    )

    df_labels_train_no_nan = handle_missing_label_values_in_df(
        df=df_labels_train,
        cat_label_columns=tabular_info.cat_columns,
        con_label_columns=tabular_info.con_columns,
        con_manual_values=train_con_means,
        name="train df",
        impute_missing=impute_missing,
    )

    df_labels_valid_no_nan = handle_missing_label_values_in_df(
        df=df_labels_valid,
        cat_label_columns=tabular_info.cat_columns,
        con_label_columns=tabular_info.con_columns,
        con_manual_values=train_con_means,
        name="valid df",
        impute_missing=impute_missing,
    )

    df_labels_full = pl.concat([df_labels_train_no_nan, df_labels_valid_no_nan])

    label_columns: al_target_columns = {
        "con": list(tabular_info.con_columns),
        "cat": list(tabular_info.cat_columns),
    }
    fit_label_transformers = _get_fit_label_transformers(
        df_labels_train=df_labels_train_no_nan,
        df_labels_full=df_labels_full,
        label_columns=label_columns,
        impute_missing=impute_missing,
    )

    logger.debug("Transforming label columns in train and validation dataframes.")
    df_train_final = transform_label_df(
        df_labels=df_labels_train_no_nan,
        label_transformers=fit_label_transformers,
        impute_missing=impute_missing,
    )
    df_valid_final = transform_label_df(
        df_labels=df_labels_valid_no_nan,
        label_transformers=fit_label_transformers,
        impute_missing=impute_missing,
    )

    return df_train_final, df_valid_final, fit_label_transformers


def _get_con_manual_vals_dict(
    df: pl.DataFrame,
    con_columns: Sequence[str],
) -> dict[str, float]:
    con_means_dict: dict[str, float] = {}

    for column in con_columns:
        column_mean = (
            df.get_column(column).drop_nans().drop_nulls().cast(pl.Float32).mean()
        )
        if isinstance(column_mean, (int, float)):
            con_means_dict[column] = float(column_mean)
        else:
            con_means_dict[column] = 0.0

    return con_means_dict


def handle_missing_label_values_in_df(
    df: pl.DataFrame,
    cat_label_columns: Sequence[str],
    con_label_columns: Sequence[str],
    impute_missing: bool,
    con_manual_values: Optional[dict[str, float]] = None,
    name: str = "df",
) -> pl.DataFrame:
    df_filled_cat = _fill_categorical_nans(
        df=df,
        column_names=cat_label_columns,
        name=name,
        impute_missing=impute_missing,
    )

    if con_manual_values is None:
        df_filled_final = df_filled_cat
    else:
        df_filled_final = _fill_continuous_nans(
            df=df_filled_cat,
            column_names=con_label_columns,
            name=name,
            con_means_dict=con_manual_values,
            impute_missing=impute_missing,
        )

    return df_filled_final


def _fill_categorical_nans(
    df: pl.DataFrame,
    column_names: Sequence[str],
    impute_missing: bool,
    name: str = "df",
) -> pl.DataFrame:
    missing_stats = _get_missing_stats_string(df=df, columns_to_check=column_names)

    if not impute_missing:
        return df

    logger.debug(
        "Replacing NaNs in categorical columns %s (counts: %s) "
        "in %s with '__NULL__'.",
        column_names,
        missing_stats,
        name,
    )

    expr = [
        pl.when(pl.col(col).is_null())
        .then(pl.lit("__NULL__"))
        .otherwise(pl.col(col))
        .alias(col)
        for col in column_names
    ]

    other_cols = [col for col in df.columns if col not in column_names]
    expr.extend([pl.col(col) for col in other_cols])

    return df.with_columns(expr)


def _fill_continuous_nans(
    df: pl.DataFrame,
    column_names: Sequence[str],
    con_means_dict: Dict[str, float],
    impute_missing: bool,
    name: str = "df",
) -> pl.DataFrame:
    missing_stats = _get_missing_stats_string(df=df, columns_to_check=column_names)

    if not impute_missing:
        return df

    logger.debug(
        "Replacing NaNs in continuous columns %s (counts: %s) in %s with %s",
        column_names,
        missing_stats,
        name,
        con_means_dict,
    )

    cols_to_fill = con_means_dict.keys()

    expr = [
        pl.when(pl.col(col).is_null() | pl.col(col).is_nan())
        .then(pl.lit(con_means_dict[col]))
        .otherwise(pl.col(col))
        .alias(col)
        for col in cols_to_fill
    ]

    other_cols = [col for col in df.columns if col not in column_names]
    expr.extend([pl.col(col) for col in other_cols])

    return df.with_columns(expr)


def _get_missing_stats_string(
    df: pl.DataFrame, columns_to_check: Sequence[str]
) -> Dict[str, int]:
    missing_count_dict = {
        col: int(df.get_column(col).is_null().sum()) for col in columns_to_check
    }
    return missing_count_dict


def merge_target_columns(
    target_con_columns: List[str], target_cat_columns: List[str]
) -> al_target_columns:
    if len(target_con_columns + target_cat_columns) == 0:
        raise ValueError("Expected at least 1 label column")

    all_target_columns: al_target_columns = {
        "con": target_con_columns,
        "cat": target_cat_columns,
    }

    assert len(all_target_columns) > 0

    return all_target_columns



import math
from typing import Callable, Optional

import torch
from torch import nn

from eir.models.layers.attention_layers import SwiGLU


class Transpose(nn.Module):
    def __init__(self, dim0, dim1):
        super().__init__()
        self.dim0 = dim0
        self.dim1 = dim1

    def forward(self, x):
        return x.transpose(self.dim0, self.dim1)


class SequenceProjectionLayer(nn.Module):
    def __init__(
        self,
        input_shape_no_batch: torch.Size,
        target_embedding_dim: int,
        target_seq_len: Optional[int] = None,
    ):
        super().__init__()
        self.input_shape = input_shape_no_batch
        self.target_embedding_dim = target_embedding_dim
        self.target_seq_len = target_seq_len
        self.reshape_func, _ = get_reshape_to_attention_dims_func(
            input_shape=input_shape_no_batch
        )
        self.projection = self._create_projection()

    def _create_projection(self):
        n_input_dims = len(self.input_shape)
        target_embedding_dim = self.target_embedding_dim
        if n_input_dims == 1:
            input_embedding_dim = self.input_shape[0]
        elif n_input_dims == 2:
            input_embedding_dim = self.input_shape[1]
        else:
            input_embedding_dim = math.prod(self.input_shape[1:])

        layers: list[nn.Module]
        layers = [
            nn.Linear(
                in_features=input_embedding_dim,
                out_features=target_embedding_dim,
            ),
        ]

        if self.target_seq_len is not None:
            if n_input_dims == 2:
                linear_layer = nn.Linear(
                    in_features=1,
                    out_features=self.target_seq_len,
                )
            elif n_input_dims > 2 and self.input_shape[1] != self.target_seq_len:
                linear_layer = nn.Linear(
                    in_features=self.input_shape[1],
                    out_features=self.target_seq_len,
                )
            else:
                return nn.Sequential(*layers)

            # here operating with batch dim, hence offset by 1 compared to the logic
            # above to get 1 and 2
            layers.extend(
                [
                    nn.RMSNorm(normalized_shape=target_embedding_dim),
                    SwiGLU(
                        in_features=target_embedding_dim,
                        hidden_features=target_embedding_dim,
                        out_features=target_embedding_dim,
                        bias=False,
                    ),
                    Transpose(1, 2),
                    linear_layer,
                    Transpose(1, 2),
                ]
            )

        return nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out = self.reshape_func(x)
        out = self.projection(out)
        return out


def get_reshape_to_attention_dims_func(
    input_shape: torch.Size,
) -> tuple[Callable[[torch.Tensor], torch.Tensor], torch.Size]:
    n_input_dims = len(input_shape)

    if n_input_dims == 1:

        def func(x):
            return x.unsqueeze(0)

        output_shape = torch.Size([1, input_shape[0]])
    elif n_input_dims == 2:

        def func(x):
            return x

        output_shape = input_shape
    else:

        def func(x):
            """
            Note here start at 2 since we have the batch dim when this is called.
            """
            return x.flatten(start_dim=2)

        output_shape = torch.Size(
            [input_shape[0], int(torch.prod(torch.tensor(input_shape[1:])).item())]
        )

    return func, output_shape



import math
from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal, Union

from torch import Tensor, nn

from eir.models.input.array.models_locally_connected import FlattenFunc
from eir.models.input.sequence.transformer_models import (
    get_positional_representation_class,
    parse_dim_feedforward,
)
from eir.models.layers.attention_layers import Transformer

if TYPE_CHECKING:
    from eir.setup.input_setup_modules.common import DataDimensions


@dataclass
class ArrayTransformerConfig:
    """
    :param patch_size:
        Controls the size of the patches used in the first layer. If set to ``None``,
        the input is flattened according to the torch ``flatten`` function. Note that
        when using this parameter, we generally want the kernel width to be set to
        the multiplication of the patch size. Order follows PyTorch convention, i.e.,
        [channels, height, width].

    :param embedding_dim:
        The embedding dimension each patch is projected to. This is also the
        dimension of the transformer encoder layers.

    :param num_heads:
        The number of heads in the multi-head attention layers.

    :param num_layers:
        The number of transformer encoder layers.

    :param dim_feedforward:
        The dimension of the feedforward layers in the transformer model.

    :param dropout:
        The dropout rate to use in the transformer encoder layers.

    :param position:
        Whether to encode the token position or use learnable position embeddings.

    :param position_dropout:
        The dropout rate to use in the position encoding/embedding.
    """

    patch_size: tuple[int, int, int]
    embedding_dim: int

    num_heads: int = 8
    num_layers: int = 2
    dim_feedforward: Union[int, Literal["auto"]] = "auto"
    dropout: float = 0.10

    position: Literal["encode", "embed"] = "encode"
    position_dropout: float = 0.10


class ArrayTransformer(nn.Module):
    def __init__(
        self,
        model_config: ArrayTransformerConfig,
        data_dimensions: "DataDimensions",
        flatten_fn: FlattenFunc,
    ) -> None:
        super().__init__()

        self.model_config = model_config
        self.data_dimensions = data_dimensions
        self.flatten_fn = flatten_fn
        self.embedding_dim = self.model_config.embedding_dim

        n_elements_per_patch = _compute_num_patch_elements(
            patch_size=self.model_config.patch_size
        )

        self.num_patches = _compute_num_patches(
            input_num_elements=self.data_dimensions.num_elements(),
            n_elements_per_patch=n_elements_per_patch,
        )

        self.embedding_projection = nn.Linear(
            in_features=n_elements_per_patch,
            out_features=self.embedding_dim,
        )

        pos_repr_class = get_positional_representation_class(
            position_model_config=self.model_config.position
        )
        self.pos_representation = pos_repr_class(
            embedding_dim=self.embedding_dim,
            dropout=self.model_config.position_dropout,
            max_length=self.num_patches,
        )

        dim_feed_forward = parse_dim_feedforward(
            dim_feedforward=self.model_config.dim_feedforward,
            embedding_dim=self.embedding_dim,
        )

        self.transformer_encoder = Transformer(
            d_model=self.embedding_dim,
            nhead=self.model_config.num_heads,
            num_layers=self.model_config.num_layers,
            dim_feedforward=dim_feed_forward,
            dropout=self.model_config.dropout,
            norm_first=True,
        )

        self.output_shape = (1, self.num_patches, self.embedding_dim)

    @property
    def num_out_features(self) -> int:
        return self.num_patches * self.embedding_dim

    def forward(self, input: Tensor) -> Tensor:
        out = self.flatten_fn(input)
        out = out.reshape(out.shape[0], self.num_patches, -1)

        out = self.embedding_projection(out)

        out = out * math.sqrt(self.embedding_dim)
        out = self.pos_representation(out)
        out = self.transformer_encoder(out)
        out = out.flatten(1)
        return out


def _compute_num_patches(input_num_elements: int, n_elements_per_patch: int) -> int:
    num_patches = int(input_num_elements / n_elements_per_patch)
    return num_patches


def _compute_num_patch_elements(patch_size: tuple[int, ...]) -> int:
    number_of_patches = 1
    for element in patch_size:
        number_of_patches *= element
    return number_of_patches



import multiprocessing
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Callable, List, Optional, Tuple, Union

import torch
import torch.multiprocessing

multiprocessing.set_start_method("spawn", force=True)

torch.multiprocessing.set_sharing_strategy("file_system")
from aislib.misc_utils import ensure_path_exists
from ignite.engine import Engine
from torch import nn
from torch.optim.optimizer import Optimizer
from torch.utils.data import DataLoader, DistributedSampler, WeightedRandomSampler

from eir import __version__
from eir.data_load import datasets
from eir.data_load.data_utils import get_train_sampler
from eir.data_load.label_setup import split_ids
from eir.experiment_io.experiment_io import get_version_file
from eir.experiment_io.input_object_io import serialize_chosen_input_objects
from eir.experiment_io.output_object_io import serialize_output_objects
from eir.models.model_setup import get_model
from eir.models.model_setup_modules.meta_setup import al_meta_model
from eir.models.model_training_utils import run_lr_find
from eir.setup.config import Configs, get_configs
from eir.setup.input_setup import al_input_objects_as_dict, set_up_inputs_for_training
from eir.setup.input_setup_modules.setup_tabular import serialize_all_input_transformers
from eir.setup.output_setup import (
    al_output_objects_as_dict,
    set_up_outputs_for_training,
)
from eir.setup.output_setup_modules.sequence_output_setup import (
    converge_sequence_input_and_output,
)
from eir.setup.streaming_data_setup.streaming_data_adapters import (
    patch_configs_for_local_data,
    setup_and_gather_streaming_data,
)
from eir.setup.streaming_data_setup.streaming_data_utils import send_validation_ids
from eir.target_setup.target_label_setup import (
    gather_all_ids_from_output_configs,
    read_manual_ids_if_exist,
    set_up_all_targets_wrapper,
)
from eir.train_utils import distributed, utils
from eir.train_utils.criteria import al_criteria_dict, get_criteria, get_loss_callable
from eir.train_utils.metrics import get_average_history_filepath, get_default_metrics
from eir.train_utils.optim import get_optimizer, maybe_wrap_model_with_swa
from eir.train_utils.step_logic import (
    Hooks,
    al_training_labels_target,
    get_default_hooks,
)
from eir.train_utils.train_handlers import configure_trainer
from eir.train_utils.utils import call_hooks_stage_iterable
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.train_utils.metrics import al_metric_record_dict, al_step_metric_dict


utils.seed_everything()
logger = get_logger(name=__name__, tqdm_compatible=True)


def main():
    torch.backends.cudnn.benchmark = True

    configs = get_configs()

    configs, local_rank = distributed.maybe_initialize_distributed_environment(
        configs=configs
    )

    default_hooks = get_default_hooks(configs=configs)
    default_experiment = get_default_experiment(
        configs=configs,
        hooks=default_hooks,
    )

    run_experiment(experiment=default_experiment)


@dataclass(frozen=True)
class Experiment:
    configs: Configs
    inputs: al_input_objects_as_dict
    outputs: al_output_objects_as_dict
    train_loader: torch.utils.data.DataLoader
    valid_loader: torch.utils.data.DataLoader
    valid_dataset: datasets.al_local_datasets
    model: al_meta_model
    optimizer: Optimizer
    criteria: al_criteria_dict
    loss_function: Callable
    metrics: "al_metric_record_dict"
    hooks: Hooks


def get_default_experiment(
    configs: Configs,
    hooks: Hooks,
) -> "Experiment":

    gc = configs.global_config

    output_folder = gc.be.output_folder
    run_folder = _prepare_run_folder(output_folder=output_folder)

    streaming_data = setup_and_gather_streaming_data(
        configs=configs,
        output_folder=output_folder,
        batch_size=gc.be.batch_size,
        # TODO: Make this configurable
        max_samples=5000,
    )

    if streaming_data is not None:
        ws_url, streaming_local_folder = streaming_data
        configs_original, configs = patch_configs_for_local_data(
            configs=configs,
            local_data_path=streaming_local_folder,
        )
    else:
        ws_url = None

    all_ids = gather_all_ids_from_output_configs(output_configs=configs.output_configs)
    manual_valid_ids = read_manual_ids_if_exist(
        manual_valid_ids_file=gc.be.manual_valid_ids_file
    )

    train_ids, valid_ids = split_ids(
        ids=all_ids,
        valid_size=gc.be.valid_size,
        manual_valid_ids=manual_valid_ids,
    )

    if ws_url is not None:
        send_validation_ids(
            ws_url=ws_url,
            valid_ids=list(valid_ids),
        )

    target_labels = set_up_all_targets_wrapper(
        train_ids=train_ids,
        valid_ids=valid_ids,
        run_folder=run_folder,
        output_configs=configs.output_configs,
        hooks=hooks,
    )

    inputs_as_dict = set_up_inputs_for_training(
        inputs_configs=configs.input_configs,
        train_ids=train_ids,
        valid_ids=valid_ids,
        hooks=hooks,
    )

    serialize_all_input_transformers(inputs_dict=inputs_as_dict, run_folder=run_folder)
    serialize_chosen_input_objects(inputs_dict=inputs_as_dict, run_folder=run_folder)

    outputs_as_dict = set_up_outputs_for_training(
        output_configs=configs.output_configs,
        input_objects=inputs_as_dict,
        target_transformers=getattr(target_labels, "label_transformers", None),
    )
    serialize_output_objects(output_objects=outputs_as_dict, run_folder=run_folder)

    inputs_as_dict = converge_sequence_input_and_output(
        inputs=inputs_as_dict, outputs=outputs_as_dict
    )

    train_dataset, valid_dataset = datasets.set_up_datasets_from_configs(
        configs=configs,
        target_labels=target_labels,
        inputs_as_dict=inputs_as_dict,
        outputs_as_dict=outputs_as_dict,
        train_ids_to_keep=train_ids,
        valid_ids_to_keep=valid_ids,
        websocket_url=ws_url,
    )

    train_sampler = None
    if isinstance(train_dataset, (datasets.DiskDataset, datasets.MemoryDataset)):
        train_sampler = get_train_sampler(
            columns_to_sample=gc.tc.weighted_sampling_columns,
            train_dataset=train_dataset,
        )

    train_dataloader, valid_dataloader = get_dataloaders(
        train_dataset=train_dataset,
        train_sampler=train_sampler,
        valid_dataset=valid_dataset,
        batch_size=gc.be.batch_size,
        num_workers=gc.be.dataloader_workers,
    )

    model = get_model(
        global_config=gc,
        inputs_as_dict=inputs_as_dict,
        fusion_config=configs.fusion_config,
        outputs_as_dict=outputs_as_dict,
    )

    model = maybe_wrap_model_with_swa(
        n_iter_before_swa=gc.m.n_iter_before_swa,
        model=model,
        device=torch.device(gc.be.device),
    )
    _log_model(
        model=model,
        structure_file=run_folder / "model_info.txt",
    )

    criteria = get_criteria(
        outputs_as_dict=outputs_as_dict,
    )

    loss_func = get_loss_callable(criteria=criteria)

    extra_modules = None
    if hooks.extra_state is not None:
        extra_modules = hooks.extra_state.get("uncertainty_modules", {})

    optimizer = get_optimizer(
        model=model,
        loss_callable=loss_func,
        global_config=gc,
        extra_modules=extra_modules,
    )

    metrics = get_default_metrics(
        target_transformers=target_labels.label_transformers,
        cat_metrics=gc.met.cat_metrics,
        con_metrics=gc.met.con_metrics,
        cat_averaging_metrics=gc.met.cat_averaging_metrics,
        con_averaging_metrics=gc.met.con_averaging_metrics,
        output_configs=configs.output_configs,
    )

    experiment = Experiment(
        configs=configs,
        inputs=inputs_as_dict,
        outputs=outputs_as_dict,
        train_loader=train_dataloader,
        valid_loader=valid_dataloader,
        valid_dataset=valid_dataset,
        model=model,
        optimizer=optimizer,
        criteria=criteria,
        loss_function=loss_func,
        metrics=metrics,
        hooks=hooks,
    )

    return experiment


def _prepare_run_folder(output_folder: str) -> Path:
    run_folder = utils.get_run_folder(output_folder=output_folder)
    history_file = get_average_history_filepath(
        run_folder=run_folder, train_or_val_target_prefix="train_"
    )
    if history_file.exists():
        raise FileExistsError(
            f"There already exists a run with that name: {history_file}. Please choose "
            f"a different run name or delete the folder."
        )

    ensure_path_exists(path=run_folder, is_folder=True)

    return run_folder


def get_dataloaders(
    train_dataset: datasets.al_datasets,
    train_sampler: Union[None, WeightedRandomSampler, DistributedSampler],
    valid_dataset: datasets.al_local_datasets,
    batch_size: int,
    num_workers: int = 0,
) -> Tuple:

    train_num_workers = num_workers

    if isinstance(train_dataset, (datasets.DiskDataset, datasets.MemoryDataset)):
        check_dataset_and_batch_size_compatibility(
            dataset=train_dataset,
            batch_size=batch_size,
            name="Training",
        )
    else:
        if num_workers > 0:
            logger.warning(
                "When using a streaming dataset with multiple "
                " workers (num_workers > 0), "
                "each worker will create its own "
                "connection to the data server. This can "
                "potentially lead to duplicate data "
                "being processed if the server is not "
                "properly configured to handle multiple connections. "
                "Ensure that your "
                "data server implements proper coordination "
                "to distribute unique samples "
                "across all connections. If you're unsure about your server's "
                "implementation, consider using num_workers=0 or consult your data "
                "server's documentation."
            )

    check_dataset_and_batch_size_compatibility(
        dataset=valid_dataset,
        batch_size=batch_size,
        name="Validation",
    )

    shuffle: Optional[bool] = False if train_sampler else True
    if isinstance(train_dataset, datasets.StreamingDataset):
        shuffle = None

    pin_memory = torch.cuda.is_available()
    if pin_memory:
        logger.debug("Enabling memory pinning for CPU -> GPU transfer.")

    persistent_workers = False
    if num_workers > 0:
        persistent_workers = True
        logger.debug("Enabling persistent workers for dataloader.")

    train_dataloader = DataLoader(
        dataset=train_dataset,
        batch_size=batch_size,
        sampler=train_sampler,
        shuffle=shuffle,
        num_workers=train_num_workers,
        pin_memory=pin_memory,
        drop_last=True,
        persistent_workers=persistent_workers,
    )

    valid_dataloader = DataLoader(
        dataset=valid_dataset,
        batch_size=batch_size,
        shuffle=False,
        num_workers=num_workers,
        pin_memory=pin_memory,
        drop_last=False,
        persistent_workers=persistent_workers,
    )

    return train_dataloader, valid_dataloader


def check_dataset_and_batch_size_compatibility(
    dataset: datasets.DatasetBase, batch_size: int, name: str = ""
):
    if len(dataset) < batch_size:
        raise ValueError(
            f"{name} dataset size ({len(dataset)}) can not be smaller than "
            f"batch size ({batch_size}). A fix can be increasing {name.lower()} sample "
            f"size or reducing the batch size. If predicting on few unknown samples, "
            f"a solution can be setting the batch size to 1 in the global configuration"
            f" passed to the prediction module. Future work includes making this "
            f"easier to work with."
        )


def _log_model(
    model: nn.Module,
    structure_file: Optional[Path],
    verbose: bool = False,
    parameter_file: Optional[str] = None,
) -> None:
    no_params = sum(p.numel() for p in model.parameters())
    no_trainable_params = sum(p.numel() for p in model.parameters() if p.requires_grad)

    model_summary = "Starting training with following model specifications:\n"
    model_summary += f"Total parameters: {format(no_params, ',.0f')}\n"
    model_summary += f"Trainable parameters: {format(no_trainable_params, ',.0f')}"

    logger.info(model_summary)

    if structure_file:
        with open(structure_file, "w") as structure_handle:
            structure_handle.write(str(model))

    layer_summary = ""
    if verbose:
        layer_summary = "\nModel layers:\n"
        for name, param in model.named_parameters():
            layer_summary += (
                f"Layer: {name}, "
                f"Shape: {list(param.size())}, "
                f"Parameters: {param.numel()}, "
                f"Trainable: {param.requires_grad}\n"
            )
        logger.info(layer_summary)

    if parameter_file:
        with open(parameter_file, "w") as f:
            f.write(model_summary)
            if verbose:
                f.write(layer_summary)


def _log_eir_version_info(outfile: Path) -> None:
    eir_version_info = f"{__version__}\n"

    ensure_path_exists(path=outfile)
    with open(outfile, "w") as f:
        f.write(eir_version_info)


def run_experiment(experiment: Experiment) -> None:
    gc = experiment.configs.global_config
    run_folder = utils.get_run_folder(output_folder=gc.be.output_folder)

    _log_eir_version_info(outfile=get_version_file(run_folder=run_folder))

    train(experiment=experiment)


def train(experiment: Experiment) -> None:
    exp = experiment
    gc = experiment.configs.global_config

    trainer = get_base_trainer(experiment=experiment)

    if gc.lr.find_lr:
        logger.info("Running LR find and exiting.")
        run_lr_find(
            trainer_engine=trainer,
            train_dataloader=exp.train_loader,
            model=exp.model,
            optimizer=exp.optimizer,
            output_folder=utils.get_run_folder(output_folder=gc.be.output_folder),
        )
        return

    trainer = configure_trainer(trainer=trainer, experiment=experiment)

    trainer.run(data=exp.train_loader, max_epochs=gc.be.n_epochs)


def get_base_trainer(experiment: Experiment) -> Engine:
    step_hooks = experiment.hooks.step_func_hooks

    def step(
        engine: Engine,
        loader_batch: Tuple[torch.Tensor, al_training_labels_target, List[str]],
    ) -> "al_step_metric_dict":
        """
        The output here goes to trainer.output.
        """
        experiment.model.train()
        experiment.optimizer.zero_grad()

        base_prepare_inputs_stage = step_hooks.base_prepare_batch
        state = call_hooks_stage_iterable(
            hook_iterable=base_prepare_inputs_stage,
            common_kwargs={"experiment": experiment, "loader_batch": loader_batch},
            state=None,
        )
        base_batch = state["batch"]

        post_prepare_inputs_stage = step_hooks.post_prepare_batch
        state = call_hooks_stage_iterable(
            hook_iterable=post_prepare_inputs_stage,
            common_kwargs={"experiment": experiment, "loader_batch": base_batch},
            state=state,
        )
        batch = state["batch"]

        model_forward_loss_stage = step_hooks.model_forward
        state = call_hooks_stage_iterable(
            hook_iterable=model_forward_loss_stage,
            common_kwargs={"experiment": experiment, "batch": batch},
            state=state,
        )

        loss_stage = step_hooks.loss
        state = call_hooks_stage_iterable(
            hook_iterable=loss_stage,
            common_kwargs={"experiment": experiment, "batch": batch},
            state=state,
        )

        optimizer_backward_stage = step_hooks.optimizer_backward
        state = call_hooks_stage_iterable(
            hook_iterable=optimizer_backward_stage,
            common_kwargs={"experiment": experiment, "batch": batch},
            state=state,
        )

        metrics_stage = step_hooks.metrics
        state = call_hooks_stage_iterable(
            hook_iterable=metrics_stage,
            common_kwargs={"experiment": experiment, "batch": batch},
            state=state,
        )

        return state["metrics"]

    trainer = Engine(process_function=step)

    return trainer


if __name__ == "__main__":
    main()



import warnings
from functools import lru_cache
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generator, Set, Union

warnings.filterwarnings("ignore", message=".*newer version of deeplake.*")

import deeplake
import numpy as np
import polars as pl
from tqdm import tqdm

if TYPE_CHECKING:
    from eir.data_load.data_preparation_modules.input_preparation_wrappers import (
        InputHookOutput,
    )


@lru_cache()
def is_deeplake_dataset(data_source: str) -> bool:
    if deeplake.exists(data_source):
        return True
    return False


@lru_cache()
def load_deeplake_dataset(data_source: str) -> deeplake.ReadOnlyDataset:
    dataset = deeplake.open_read_only(
        url=data_source,
        token=None,
    )

    columns = {col.name for col in dataset.schema.columns}
    if "ID" not in columns:
        raise ValueError(
            f"DeepLake dataset at {data_source} does not have an ID column. "
            f"Please add one to the dataset."
        )
    return dataset


def get_deeplake_input_source_iterable(
    deeplake_dataset: deeplake.ReadOnlyDataset, inner_key: str
) -> Generator[Union[deeplake.Column, deeplake.ColumnView], None, None]:
    columns = {col.name for col in deeplake_dataset.schema.columns}

    existence_col = f"{inner_key}_exists"
    for row in deeplake_dataset:
        if is_deeplake_sample_missing(
            row=row,
            existence_col=existence_col,
            columns=columns,
        ):
            continue

        value = row[inner_key]

        yield value


def add_deeplake_data_to_df(
    input_source: str,
    input_name: str,
    deeplake_input_inner_key: str,
    input_df: pl.DataFrame,
    data_loading_hook: "InputHookOutput",
    ids_to_keep: Union[None, Set[str]],
) -> pl.DataFrame:
    assert deeplake_input_inner_key is not None, "Deeplake input inner key is None"

    deeplake_ds = load_deeplake_dataset(data_source=input_source)

    columns = {col.name for col in deeplake_ds.schema.columns}
    if deeplake_input_inner_key not in columns:
        raise ValueError(
            f"Input key {deeplake_input_inner_key} not found in deeplake dataset "
            f"{input_source}. Available columns are: {columns}."
        )

    total = len(ids_to_keep) if ids_to_keep is not None else len(deeplake_ds)
    existence_col = f"{deeplake_input_inner_key}_exists"

    ids = []
    column_arrays: dict[str, Any] = {}
    hook_callable = data_loading_hook.hook_callable
    hook_dtype = data_loading_hook.return_dtype
    is_list_dype = isinstance(hook_dtype, pl.List)

    with tqdm(total=total, desc=input_name) as pbar:
        for row in deeplake_ds:
            if is_deeplake_sample_missing(
                row=row,
                existence_col=existence_col,
                columns=columns,
            ):
                continue

            sample_id = row["ID"]

            if ids_to_keep is not None and sample_id not in ids_to_keep:
                continue

            sample_data_pointer = row.row_id
            sample_data = hook_callable(sample_data_pointer)

            if isinstance(sample_data, Path):
                sample_data = str(sample_data)

            if isinstance(sample_data, dict):
                for key in sample_data.keys():
                    col_name = f"{input_name}__{key}"
                    if col_name not in column_arrays:
                        column_arrays[col_name] = []

                for key, value in sample_data.items():
                    col_name = f"{input_name}__{key}"
                    column_arrays[col_name].append(value)
                ids.append(sample_id)
            else:
                if input_name not in column_arrays:
                    column_arrays[input_name] = []

                if is_list_dype and isinstance(sample_data, np.ndarray):
                    sample_data = sample_data.tolist()

                column_arrays[input_name].append(sample_data)
                ids.append(sample_id)

            pbar.update(1)

    if not ids:
        return input_df

    df_dict = {"ID": pl.Series(name="ID", values=ids, dtype=pl.Utf8)}

    for col_name, values in column_arrays.items():
        df_dict[col_name] = pl.Series(name=col_name, values=values, dtype=hook_dtype)

    processed_df = pl.DataFrame(df_dict)

    if input_df.height == 0:
        return processed_df
    else:
        return input_df.join(processed_df, on="ID", how="full", coalesce=True)


def is_deeplake_sample_missing(
    row: deeplake.RowView,
    existence_col: str,
    columns: set[str],
) -> bool:
    if existence_col in columns:
        is_missing = not row[existence_col].item()
        return is_missing

    return False



from copy import copy
from typing import Any, Dict, Literal, Type, Union

from eir.models.input.array.models_locally_connected import LCLModelConfig
from eir.models.output.array.array_output_modules import (
    ArrayOutputModuleConfig,
    LCLOutputModelConfig,
)
from eir.models.output.array.output_array_models_cnn import CNNUpscaleModelConfig
from eir.models.output.sequence.sequence_output_modules import (
    SequenceOutputModuleConfig,
    TransformerSequenceOutputModuleConfig,
)
from eir.models.output.tabular.linear import LinearOutputModuleConfig
from eir.models.output.tabular.mlp_residual import ResidualMLPOutputModuleConfig
from eir.models.output.tabular.shared_mlp_residual import (
    SharedResidualMLPOutputModuleConfig,
)
from eir.models.output.tabular.tabular_output_modules import TabularOutputModuleConfig
from eir.setup import schemas
from eir.setup.config_setup_modules.config_setup_utils import (
    validate_keys_against_dataclass,
)
from eir.setup.tensor_broker_setup import set_up_tensor_broker_config
from eir.utils.logging import get_logger

al_output_model_config_classes = (
    Type[ResidualMLPOutputModuleConfig]
    | Type[LinearOutputModuleConfig]
    | Type[SharedResidualMLPOutputModuleConfig]
    | Type[TransformerSequenceOutputModuleConfig]
    | Type[LCLModelConfig]
    | Type[CNNUpscaleModelConfig]
)
al_output_model_configs = (
    ResidualMLPOutputModuleConfig
    | LinearOutputModuleConfig
    | SharedResidualMLPOutputModuleConfig
    | TransformerSequenceOutputModuleConfig
    | LCLModelConfig
    | CNNUpscaleModelConfig
)
al_output_model_init_map = dict[str, dict[str, al_output_model_config_classes]]

logger = get_logger(name=__name__)


def init_output_config(
    yaml_config_as_dict: Dict[str, Any],
) -> schemas.OutputConfig:
    cfg = yaml_config_as_dict

    validate_keys_against_dataclass(
        input_dict=cfg,
        dataclass_type=schemas.OutputConfig,
        name=cfg.get("output_info", {}).get("output_name", ""),
    )

    output_info_object = schemas.OutputInfoConfig(**cfg["output_info"])

    output_schema_map = get_outputs_types_schema_map()
    output_type_info_class = output_schema_map[output_info_object.output_type]

    output_type_info_class_init_kwargs = cfg.get("output_type_info", {})
    output_type_info_object = output_type_info_class(
        **output_type_info_class_init_kwargs
    )

    model_config = set_up_output_module_config(
        output_info_object=output_info_object,
        model_init_kwargs_base=cfg.get("model_config", {}),
    )

    sampling_config = _set_up_basic_sampling_config(
        output_type_config=output_type_info_object,
        sampling_config=cfg.get("sampling_config", {}),
    )

    tensor_broker_config = set_up_tensor_broker_config(
        tensor_broker_config=cfg.get("tensor_broker_config", {})
    )

    output_config = schemas.OutputConfig(
        output_info=output_info_object,
        output_type_info=output_type_info_object,
        model_config=model_config,
        sampling_config=sampling_config,
        tensor_broker_config=tensor_broker_config,
    )

    return output_config


def _set_up_basic_sampling_config(
    output_type_config: schemas.al_output_type_configs, sampling_config: dict
) -> dict | schemas.ArrayOutputSamplingConfig | schemas.ImageOutputSamplingConfig:
    """
    Note that the sequence sampling config currently has it's own logic
    in output_config_setup_sequence.py.
    """
    sampling_config_object: (
        dict | schemas.ArrayOutputSamplingConfig | schemas.ImageOutputSamplingConfig
    )
    match output_type_config:
        case schemas.ArrayOutputTypeConfig():
            sampling_config_object = schemas.ArrayOutputSamplingConfig(
                **sampling_config
            )
        case schemas.ImageOutputTypeConfig():
            sampling_config_object = schemas.ImageOutputSamplingConfig(
                **sampling_config
            )

        case (
            schemas.TabularOutputTypeConfig()
            | schemas.SequenceOutputTypeConfig()
            | schemas.SurvivalOutputTypeConfig()
        ):
            sampling_config_object = sampling_config
        case _:
            raise ValueError(f"Unknown output type config '{output_type_config}'.")

    return sampling_config_object


def get_outputs_types_schema_map() -> Dict[
    str,
    Type[schemas.TabularOutputTypeConfig]
    | Type[schemas.SequenceOutputTypeConfig]
    | Type[schemas.ArrayOutputTypeConfig]
    | Type[schemas.ImageOutputTypeConfig]
    | Type[schemas.SurvivalOutputTypeConfig],
]:
    mapping = {
        "tabular": schemas.TabularOutputTypeConfig,
        "sequence": schemas.SequenceOutputTypeConfig,
        "array": schemas.ArrayOutputTypeConfig,
        "image": schemas.ImageOutputTypeConfig,
        "survival": schemas.SurvivalOutputTypeConfig,
    }

    return mapping


def get_output_module_config_class(
    output_type: str,
) -> schemas.al_output_module_configs_classes:
    model_config_setup_map = get_output_module_config_class_map()

    return model_config_setup_map[output_type]


def get_output_module_config_class_map() -> (
    Dict[str, schemas.al_output_module_configs_classes]
):
    mapping = {
        "tabular": TabularOutputModuleConfig,
        "sequence": SequenceOutputModuleConfig,
        "array": ArrayOutputModuleConfig,
        "image": ArrayOutputModuleConfig,
        "survival": TabularOutputModuleConfig,
    }

    return mapping


def set_up_output_module_config(
    output_info_object: schemas.OutputInfoConfig,
    model_init_kwargs_base: dict,
) -> schemas.al_output_module_configs:
    output_type = output_info_object.output_type

    model_config_class = get_output_module_config_class(output_type=output_type)

    model_type = None
    if model_init_kwargs_base:
        model_type = model_init_kwargs_base.get("model_type", None)

    if not model_type:
        try:
            model_type = getattr(model_config_class, "model_type")
        except AttributeError:
            raise AttributeError(
                "Not model type specified in model config and could not find default "
                "value for '%s'.",
                output_type,
            )

        logger.info(
            "Output model type not specified in model configuration with name '%s', "
            "attempting to grab default value.",
            output_info_object.output_name,
        )

    output_module_init_class_map = get_output_config_type_init_callable_map()

    model_type_config = set_up_output_module_init_config(
        model_init_kwargs_base=model_init_kwargs_base.get("model_init_config", {}),
        output_type=output_type,
        model_type=model_type,
        output_module_init_class_map=output_module_init_class_map,
    )

    common_kwargs = {"model_type": model_type, "model_init_config": model_type_config}
    other_specific_kwargs = {
        k: v for k, v in model_init_kwargs_base.items() if k not in common_kwargs
    }
    model_config_kwargs = {**common_kwargs, **other_specific_kwargs}
    model_config = model_config_class(**model_config_kwargs)

    return model_config


def set_up_output_module_init_config(
    model_init_kwargs_base: Union[None, dict],
    output_type: Literal["tabular", "sequence", "array"],
    model_type: str,
    output_module_init_class_map: al_output_model_init_map,
) -> al_output_model_configs:
    if not model_init_kwargs_base:
        model_init_kwargs_base = {}

    model_init_kwargs = copy(model_init_kwargs_base)

    model_init_config_callable = output_module_init_class_map[output_type][model_type]

    model_init_config = model_init_config_callable(**model_init_kwargs)

    return model_init_config


def get_output_config_type_init_callable_map() -> al_output_model_init_map:
    mapping: al_output_model_init_map = {
        "tabular": {
            "mlp_residual": ResidualMLPOutputModuleConfig,
            "linear": LinearOutputModuleConfig,
            "shared_mlp_residual": SharedResidualMLPOutputModuleConfig,
        },
        "sequence": {
            "sequence": TransformerSequenceOutputModuleConfig,
        },
        "array": {
            "lcl": LCLOutputModelConfig,
            "cnn": CNNUpscaleModelConfig,
        },
        "image": {
            "lcl": LCLOutputModelConfig,
            "cnn": CNNUpscaleModelConfig,
        },
        "survival": {
            "mlp_residual": ResidualMLPOutputModuleConfig,
            "linear": LinearOutputModuleConfig,
            "shared_mlp_residual": SharedResidualMLPOutputModuleConfig,
        },
    }

    return mapping



from dataclasses import dataclass

from eir.data_load.label_setup import al_label_transformers
from eir.setup.schemas import InputConfig


@dataclass
class ComputedServeTabularInputInfo:
    labels: "ServeLabels"
    input_config: InputConfig


@dataclass
class ServeLabels:
    label_transformers: al_label_transformers



from functools import partial
from pathlib import Path
from typing import Callable, Dict

from eir.experiment_io.input_object_io import load_serialized_input_object
from eir.setup.input_setup_modules.setup_array import set_up_array_input_object
from eir.setup.input_setup_modules.setup_bytes import ComputedBytesInputInfo
from eir.setup.input_setup_modules.setup_image import ComputedImageInputInfo
from eir.setup.input_setup_modules.setup_omics import ComputedOmicsInputInfo
from eir.setup.input_setup_modules.setup_sequence import ComputedSequenceInputInfo
from eir.setup.input_setup_modules.setup_tabular import (
    set_up_tabular_input_from_pretrained,
)


def get_input_setup_from_pretrained_function_map(
    run_folder: Path, load_module_name: str
) -> Dict[str, Callable]:
    pretrained_setup_mapping: Dict[str, Callable] = {
        "omics": partial(
            load_serialized_input_object,
            input_class=ComputedOmicsInputInfo,
            run_folder=run_folder,
            custom_input_name=load_module_name,
        ),
        "tabular": partial(
            set_up_tabular_input_from_pretrained,
            custom_input_name=load_module_name,
        ),
        "sequence": partial(
            load_serialized_input_object,
            input_class=ComputedSequenceInputInfo,
            run_folder=run_folder,
            custom_input_name=load_module_name,
        ),
        "bytes": partial(
            load_serialized_input_object,
            input_class=ComputedBytesInputInfo,
            run_folder=run_folder,
            custom_input_name=load_module_name,
        ),
        "image": partial(
            load_serialized_input_object,
            input_class=ComputedImageInputInfo,
            run_folder=run_folder,
            custom_input_name=load_module_name,
        ),
        "array": set_up_array_input_object,
    }

    return pretrained_setup_mapping



from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Dict, Sequence, Union

from eir.data_load.label_setup import (
    Labels,
    TabularFileInfo,
    set_up_train_and_valid_tabular_data,
)
from eir.experiment_io.io_utils import get_run_folder_from_model_path
from eir.experiment_io.label_transformer_io import (
    load_transformers,
    save_transformer_set,
)
from eir.models.input.tabular.tabular import get_unique_values_from_transformers
from eir.setup import schemas

if TYPE_CHECKING:
    from eir.setup.input_setup import al_input_objects_as_dict
    from eir.train_utils.step_logic import Hooks


@dataclass
class ComputedTabularInputInfo:
    labels: Labels
    input_config: schemas.InputConfig
    total_num_features: int


def set_up_tabular_input_for_training(
    input_config: schemas.InputConfig,
    train_ids: Sequence[str],
    valid_ids: Sequence[str],
    *args,
    **kwargs,
) -> ComputedTabularInputInfo:
    input_type_info = input_config.input_type_info
    assert isinstance(input_type_info, schemas.TabularInputDataConfig)

    tabular_file_info = get_tabular_input_file_info(
        input_source=input_config.input_info.input_source,
        tabular_data_type_config=input_type_info,
    )

    tabular_labels = set_up_train_and_valid_tabular_data(
        tabular_file_info=tabular_file_info,
        train_ids=train_ids,
        valid_ids=valid_ids,
        impute_missing=True,
    )

    total_num_features = get_tabular_num_features(
        label_transformers=tabular_labels.label_transformers,
        cat_columns=list(input_type_info.input_cat_columns),
        con_columns=list(input_type_info.input_con_columns),
    )
    tabular_input_info = ComputedTabularInputInfo(
        labels=tabular_labels,
        input_config=input_config,
        total_num_features=total_num_features,
    )

    return tabular_input_info


def get_tabular_input_file_info(
    input_source: str,
    tabular_data_type_config: schemas.TabularInputDataConfig,
) -> TabularFileInfo:
    table_info = TabularFileInfo(
        file_path=Path(input_source),
        con_columns=tabular_data_type_config.input_con_columns,
        cat_columns=tabular_data_type_config.input_cat_columns,
        parsing_chunk_size=tabular_data_type_config.label_parsing_chunk_size,
    )

    return table_info


def get_tabular_num_features(
    label_transformers: Dict, cat_columns: list[str], con_columns: list[str]
) -> int:
    unique_cat_values = get_unique_values_from_transformers(
        transformers=label_transformers,
        keys_to_use=cat_columns,
    )
    cat_num_features = sum(
        len(unique_values) for unique_values in unique_cat_values.values()
    )
    total_num_features = cat_num_features + len(con_columns)

    return total_num_features


def set_up_tabular_input_from_pretrained(
    input_config: schemas.InputConfig,
    custom_input_name: str,
    train_ids: Sequence[str],
    valid_ids: Sequence[str],
    hooks: Union["Hooks", None],
) -> "ComputedTabularInputInfo":
    tabular_input_object = set_up_tabular_input_for_training(
        input_config=input_config, train_ids=train_ids, valid_ids=valid_ids, hooks=hooks
    )

    pretrained_config = input_config.pretrained_config
    assert pretrained_config is not None
    pretrained_run_folder = get_run_folder_from_model_path(
        model_path=pretrained_config.model_path
    )

    loaded_transformers = load_transformers(
        run_folder=pretrained_run_folder, transformers_to_load=None
    )
    loaded_transformers_input = loaded_transformers[custom_input_name]

    tabular_input_object.labels.label_transformers = loaded_transformers_input

    return tabular_input_object


def serialize_all_input_transformers(
    inputs_dict: "al_input_objects_as_dict", run_folder: Path
):
    for input_name, input_object in inputs_dict.items():
        match input_object:
            case ComputedTabularInputInfo(labels, _, _):
                transformers_per_source = {input_name: labels.label_transformers}
                save_transformer_set(
                    transformers_per_source=transformers_per_source,
                    run_folder=run_folder,
                )



import json
from copy import deepcopy
from pathlib import Path

from eir.experiment_io.io_utils import load_dataclass
from eir.experiment_io.output_object_io_modules.output_io_utils import (
    load_output_config_from_yaml,
)
from eir.setup import schemas
from eir.setup.output_setup_modules.image_output_setup import (
    ComputedImageOutputInfo,
    DiffusionConfig,
    ImageNormalizationStats,
    set_up_image_output,
)


def load_image_output_object(
    serialized_output_folder: Path,
) -> ComputedImageOutputInfo:
    config_path = serialized_output_folder / "output_config.yaml"
    normalization_stats_path = serialized_output_folder / "normalization_stats.json"
    num_channels_path = serialized_output_folder / "num_channels.json"
    diffusion_path = serialized_output_folder / "diffusion.json"

    output_config = load_output_config_from_yaml(output_config_path=config_path)
    output_type_info_modified = deepcopy(output_config.output_type_info)
    assert isinstance(output_type_info_modified, schemas.ImageOutputTypeConfig)

    normalization_stats = load_dataclass(
        cls=ImageNormalizationStats, file_path=normalization_stats_path
    )

    image_mode = output_type_info_modified.mode
    if not image_mode:
        num_channels = json.loads(num_channels_path.read_text())["num_channels"]
        output_type_info_modified.num_channels = num_channels
    else:
        output_type_info_modified.num_channels = None

    output_config_modified = deepcopy(output_config)
    output_config_modified.output_type_info = output_type_info_modified

    diffusion_config = None
    if diffusion_path.exists():
        diffusion_config = load_dataclass(
            cls=DiffusionConfig,
            file_path=diffusion_path,
        )

    loaded_object = set_up_image_output(
        output_config=output_config_modified,
        normalization_stats=normalization_stats,
        diffusion_config=diffusion_config,
    )

    return loaded_object



from copy import deepcopy
from pathlib import Path

from eir.experiment_io.input_object_io_modules.input_io_utils import (
    load_input_config_from_yaml,
)
from eir.experiment_io.io_utils import load_dataclass
from eir.setup.input_setup_modules.setup_omics import (
    ComputedOmicsInputInfo,
    DataDimensions,
    set_up_omics_input,
)
from eir.setup.schemas import OmicsInputDataConfig


def load_omics_input_object(serialized_input_folder: Path) -> ComputedOmicsInputInfo:
    config_path = serialized_input_folder / "input_config.yaml"
    data_dimensions_path = serialized_input_folder / "data_dimensions.json"
    snps_path = serialized_input_folder / "snps.bim"
    subset_snps_file_path = serialized_input_folder / "subset_snps_file.txt"

    input_config = load_input_config_from_yaml(input_config_path=config_path)
    input_config_copy = deepcopy(input_config)

    input_type_info = input_config.input_type_info
    assert isinstance(input_type_info, OmicsInputDataConfig)

    if snps_path.exists():
        input_type_info.snp_file = str(snps_path)

    if subset_snps_file_path.exists():
        input_type_info.subset_snps_file = str(subset_snps_file_path)

    input_config_copy.input_type_info = input_type_info

    data_dimensions = load_dataclass(
        cls=DataDimensions,
        file_path=data_dimensions_path,
    )

    loaded_object = set_up_omics_input(
        input_config=input_config_copy,
        data_dimensions=data_dimensions,
    )

    return loaded_object



from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal, Type

import torch
from torch import nn

from eir.models.input.array.array_models import al_pre_normalization
from eir.models.input.array.models_locally_connected import LCLModel, LCLModelConfig
from eir.models.layers.projection_layers import get_1d_projection_layer
from eir.models.output.array.output_array_models_cnn import (
    CNNPassThroughUpscaleModel,
    CNNUpscaleModel,
    CNNUpscaleModelConfig,
)

if TYPE_CHECKING:
    from eir.setup.input_setup_modules.common import DataDimensions

al_array_model_types = Literal["lcl", "cnn", "cnn-passthrough"]
al_output_array_model_classes = (
    Type[LCLModel] | Type[CNNUpscaleModel] | Type[CNNPassThroughUpscaleModel]
)
al_output_array_models = LCLModel | CNNUpscaleModel | CNNPassThroughUpscaleModel
al_output_array_model_config_classes = (
    Type["LCLOutputModelConfig"] | Type[CNNUpscaleModelConfig]
)


@dataclass
class LCLOutputModelConfig(LCLModelConfig):
    cutoff: int | Literal["auto"] = "auto"


@dataclass
class ArrayOutputModuleConfig:
    """
    :param model_type:
        Which type of image model to use.

    :param model_init_config:
        Configuration used to initialise model.
    """

    model_type: al_array_model_types
    model_init_config: LCLOutputModelConfig | CNNUpscaleModelConfig
    pre_normalization: al_pre_normalization = None


class ArrayOutputWrapperModule(nn.Module):
    def __init__(
        self,
        feature_extractor: al_output_array_models,
        output_name: str,
        target_data_dimensions: "DataDimensions",
    ):
        super().__init__()
        self.feature_extractor = feature_extractor
        self.output_name = output_name
        self.data_dimensions = target_data_dimensions

        self.target_width = self.data_dimensions.num_elements()
        self.target_shape = self.data_dimensions.full_shape()

        self.projection_head = get_1d_projection_layer(
            input_dimension=self.feature_extractor.num_out_features,
            target_dimension=self.target_width,
            projection_layer_type="auto",
            lcl_diff_tolerance=0,
        )

    def forward(
        self,
        x: torch.Tensor,
    ) -> dict[str, torch.Tensor]:
        out = self.feature_extractor(x)

        out = out.reshape(out.shape[0], -1)
        out = self.projection_head(out)

        out = out[:, : self.target_width]

        out = out.reshape(-1, *self.target_shape)

        return {self.output_name: out}


def get_array_output_module(
    feature_extractor: al_output_array_models,
    output_name: str,
    target_data_dimensions: "DataDimensions",
) -> ArrayOutputWrapperModule:
    return ArrayOutputWrapperModule(
        feature_extractor=feature_extractor,
        output_name=output_name,
        target_data_dimensions=target_data_dimensions,
    )



from dataclasses import dataclass
from typing import Dict, Optional

import numpy as np
from sklearn.preprocessing import KBinsDiscretizer

from eir.data_load.label_setup import (
    al_label_transformers,
    al_target_columns,
    merge_target_columns,
)
from eir.setup import schemas
from eir.setup.schema_modules.output_schemas_survival import SurvivalOutputTypeConfig
from eir.target_setup.target_setup_utils import IdentityTransformer
from eir.utils.logging import get_logger

logger = get_logger(name=__name__)

al_num_outputs_per_target = Dict[str, int]


@dataclass
class ComputedSurvivalOutputInfo:
    output_config: schemas.OutputConfig
    num_outputs_per_target: al_num_outputs_per_target
    target_columns: al_target_columns
    target_transformers: al_label_transformers

    baseline_hazard: Optional[np.ndarray] = None
    baseline_unique_times: Optional[np.ndarray] = None


def set_up_survival_output(
    output_config: schemas.OutputConfig,
    target_transformers: Dict[str, al_label_transformers],
    baseline_hazard: Optional[np.ndarray] = None,
    baseline_unique_times: Optional[np.ndarray] = None,
    *args,
    **kwargs,
) -> ComputedSurvivalOutputInfo:
    output_name = output_config.output_info.output_name
    cur_target_transformers = target_transformers[output_name]
    num_outputs_per_target = set_up_num_survival_outputs(
        target_transformers=cur_target_transformers
    )

    output_type_info = output_config.output_type_info
    assert isinstance(output_type_info, SurvivalOutputTypeConfig)

    target_columns = merge_target_columns(
        target_con_columns=[output_type_info.time_column],
        target_cat_columns=[output_type_info.event_column],
    )

    tabular_output_info = ComputedSurvivalOutputInfo(
        output_config=output_config,
        num_outputs_per_target=num_outputs_per_target,
        target_columns=target_columns,
        target_transformers=cur_target_transformers,
        baseline_hazard=baseline_hazard,
        baseline_unique_times=baseline_unique_times,
    )

    return tabular_output_info


def set_up_num_survival_outputs(
    target_transformers: al_label_transformers,
) -> al_num_outputs_per_target:
    num_outputs_per_target_dict = {}
    assert len(target_transformers) == 2
    for column, transformer in target_transformers.items():
        if not isinstance(transformer, (KBinsDiscretizer, IdentityTransformer)):
            continue

        time_column = column
        event_column = next(i for i in target_transformers if i != time_column)

        match transformer:
            case IdentityTransformer():
                num_outputs_per_target_dict[event_column] = 1
            case KBinsDiscretizer(n_bins=transformer.n_bins):
                num_outputs_per_target_dict[event_column] = transformer.n_bins

    return num_outputs_per_target_dict



from dataclasses import dataclass
from typing import TYPE_CHECKING, Dict, Literal

import torch
from torch import nn

from eir.models.fusion.fusion_attention import MetaSequenceProjection
from eir.models.input.sequence.transformer_models import (
    BasicTransformerFeatureExtractorModelConfig,
    parse_dim_feedforward,
)
from eir.models.layers.attention_layers import Transformer
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.models.model_setup_modules.meta_setup import FeatureExtractorInfo
    from eir.setup.output_setup_modules.sequence_output_setup import (
        ComputedSequenceOutputInfo,
    )

al_sequence_output_models = Literal["sequence"]

logger = get_logger(name=__name__)


@dataclass
class TransformerSequenceOutputModuleConfig(
    BasicTransformerFeatureExtractorModelConfig
):
    pass


@dataclass
class SequenceOutputModuleConfig:
    """
    :param model_init_config:
          Configuration / arguments used to initialise model.

    :param model_type:
         Which type of image model to use.

    :param embedding_dim:
        Which dimension to use for the embeddings. If ``None``, will automatically set
        this value based on the number of tokens and attention heads.

    :param position:
        Whether to encode the token position or use learnable position embeddings.

    :param position_dropout:
        Dropout for the positional encoding / embedding.

    """

    model_init_config: TransformerSequenceOutputModuleConfig
    model_type: Literal["sequence"] = "sequence"

    embedding_dim: int = 64

    position: Literal["encode", "embed"] = "encode"
    position_dropout: float = 0.10

    projection_layer_type: Literal["auto", "lcl", "lcl_residual", "linear"] = "auto"


class SequenceOutputModule(nn.Module):
    def __init__(
        self,
        output_object: "ComputedSequenceOutputInfo",
        output_name: str,
        feature_dimensionalities_and_types: Dict[str, "FeatureExtractorInfo"],
        *args,
        **kwargs,
    ):
        super().__init__()

        self.num_tokens = len(output_object.vocab)
        self.input_dimensions = feature_dimensionalities_and_types
        self.embedding_dim = output_object.embedding_dim
        self.max_length = output_object.computed_max_length
        self.output_name = output_name
        self.output_model_config = output_object.output_config.model_config
        assert isinstance(self.output_model_config, SequenceOutputModuleConfig)
        self.output_model_init_config = self.output_model_config.model_init_config
        assert isinstance(
            self.output_model_init_config, TransformerSequenceOutputModuleConfig
        )

        dim_feed_forward = parse_dim_feedforward(
            dim_feedforward=self.output_model_init_config.dim_feedforward,
            embedding_dim=self.embedding_dim,
        )

        mask = torch.triu(
            torch.ones(self.max_length, self.max_length) * float("-inf"),
            diagonal=1,
        )
        self.register_buffer("mask", mask)

        self.output_transformer = Transformer(
            d_model=self.embedding_dim,
            nhead=self.output_model_init_config.num_heads,
            num_layers=self.output_model_init_config.num_layers,
            dim_feedforward=dim_feed_forward,
            dropout=self.output_model_init_config.dropout,
            norm_first=True,
        )

        self.match_projections = nn.ModuleDict()
        for input_name, feature_extractor_info in self.input_dimensions.items():
            if input_name == self.output_name:
                continue

            match feature_extractor_info.input_type:
                case "sequence":
                    in_embed = feature_extractor_info.input_dimension.width
                case _:
                    in_embed = feature_extractor_info.output_dimension

            in_elements = feature_extractor_info.output_dimension
            cur_projection = MetaSequenceProjection(
                in_total_num_elements=in_elements,
                in_embedding_dim=in_embed,
                target_embedding_dim=self.embedding_dim,
                target_max_length=self.max_length,
                projection_layer_type=self.output_model_config.projection_layer_type,
            )

            self.match_projections[input_name] = cur_projection

        self.head = nn.Linear(self.embedding_dim, self.num_tokens)

    def forward(self, input: dict[str, torch.Tensor]) -> dict[str, torch.Tensor]:
        out = input[self.output_name]
        out = out.reshape(out.shape[0], self.max_length, self.embedding_dim)

        for input_name, input_tensor in input.items():
            if input_name == self.output_name:
                continue

            cur_projection = self.match_projections[input_name]
            projected = cur_projection(input_tensor=input_tensor, target_tensor=out)
            out = out + projected

        out = self.output_transformer(out, mask=self.mask)

        out = self.head(out)

        return {self.output_name: out}
